import unittest
from sys import argv

def verbosity() -> bool:
    return ('-v' in argv) or ('--verbose' in argv)

class FeupyTestCase(unittest.TestCase):
    """This class is unittest.TestCase plus a few utility methods, pretty much
    """
    def assertObjectAttributes(self, test_object, test_vars : dict) -> None:
        """Tests all the attributes of object specified in test_vars.
        test_vars is something akin to vars(object). Another note: test_vars
        only needs to contain the object attributes that you want to test."""
                
        for key in test_vars:
            if verbosity():
                print(f"Testing attribute '{key}'")
            self.assertEqual(getattr(test_object, key), test_vars[key])
    
    

if __name__ == '__main__':
    unittest.main()


import io
import re
from datetime import datetime, time
from functools import reduce

import bs4
import requests
from bs4 import BeautifulSoup
from lxml.html.clean import Cleaner
from PIL import Image

SIG_URLS = {
    "authentication"                      : "vld_validacao.validacao",
    "student page"                        : "fest_geral.cursos_list",
    "academic pathway"                    : "fest_geral.curso_percurso_academico_view",
    "position in the plan"                : "fest_geral.curso_posicao_plano_view",
    "courses units"                       : "fest_geral.ucurr_inscricoes_list",
    "status and more"                     : "fest_geral.estatutos_regimes_view",
    "ingress data"                        : "fest_geral.info_ingresso_view",
    "classes data"                        : "it_geral.resultado_aluno",
    "personal timetable"                  : "hor_geral.estudantes_view",
    "picture"                             : "fotografias_service.foto",
    "curricular unit"                     : "ucurr_geral.ficha_uc_view",
    "curricular unit students"            : "fest_geral.estudantes_inscritos_list",
    "curricular unit other occurrences"   : "ucurr_geral.ficha_uc_list",
    "curricular unit statistics"          : "est_geral.dist_result_ocorr",
    "curricular unit grades distribution" : "est_geral.dist_result_ocorr_detail",
    "curricular unit stats history"       : "est_geral.ucurr_result_resumo",
    "curricular unit classes"             : "it_listagem.lista_cursos_disciplina",
    "curricular unit results"             : "lres_geral.show_pautas_resul",
    "curricular unit timetable"           : "hor_geral.ucurr_view",
    "curricular unit exams"               : "exa_geral.mapa_de_exames",
    "teacher"                             : "func_geral.formview",
    "course"                              : "cur_geral.cur_view",
    "course classes"                      : "hor_geral.lista_turmas_curso",
    "course exams"                        : "exa_geral.mapa_de_exames",
    "redirection page"                    : "vld_entidades_geral.entidade_pagina"
}


def get_current_academic_year():
    today = datetime.now()
    if today.month >= 9:# 9 -> September
        return today.year
    else:
        return today.year - 1

def scrape_html_table(bs_table, f = lambda tags_list, index : list( map(lambda tag: tag.string, tags_list) )):
    """
    Returns a nested list based on the contents of the table

    html = \"\"\"<table class="formulario">
        <tr>
            <td class="very important stuff">Hotel</td>
            <td>Trivago</td>
        </tr>
        <tr>
            <td class="the class name doesn't matter">A</td>
            <td>B</td>
            <td>C</td>
        </tr>
    </table>\"\"\"
    table_soup = BeautifulSoup(html, 'lxml').table #if it's not a table, you'll get a TypeError
    print(scrape_html_table(table_soup))
    #[['Hotel', 'Trivago'], ['A', 'B', 'C']]

    There is an optional second argument that lets you define a function that operates over each row of the table(tags_list),
    which is internally represented as a list of BeautifulSoup "th" and "td" tags. The second argument of the funtion 
    is an integer that represents the index of the table row (starts at 0)

    #ex1: do nothing
    print(scrape_html_table(table_soup, lambda tags_list, index : tags_list))
    #[[<td class="very important stuff">Hotel</td>, <td>Trivago</td>], [<td class="the class name doesn't matter">A</td>, <td>B</td>, <td>C</td>]]

    #ex2: make a list of the first column of the table
    print(scrape_html_table(table_soup, lambda tags_list, index : tags_list[0].string))
    #['Hotel', 'A']

    #ex3: make a list of the tags' names
    print(scrape_html_table(table_soup, lambda tags_list, index : list(map(lambda tag : tag.name, tags_list))))
    #[['td', 'td'], ['td', 'td', 'td']]

    Don't forget that you can use the capabilities of the BeautifulSoup library to do whatever you want with these tag objects
    """
    if type(bs_table) != bs4.element.Tag:
        raise TypeError("scrape_html_table() 'bs_table' argument must be a bs4.element.Tag, not '{0}'".format( type(bs_table).__name__ ))

    if type(f).__name__ != "function": # Look, it works:/
        raise TypeError("scrape_html_table() 'f' argument must be a function, not '{0}'".format( type(f).__name__ ))
    
    if bs_table.name != "table" and bs_table.name != "thead" and bs_table.name != "tbody":
        raise ValueError("The tag element must be a table, not '{0}'".format(bs_table.name))

    result = []
    for index, tr in enumerate(bs_table.find_all("tr", recursive = False)):  # split the table into table rows
        tags = tr.find_all(["th", "td"], recursive = False)                  # split the table row into table header and data tags
        result.append(f(tags, index))                                        # apply f to the tags list and append to result
    
    return result

def trim_html(html):
    """Takes a html string as input and returns the html without any styles nor javascript"""
    cleaner = Cleaner()

    cleaner.scripts         = True
    cleaner.javascript      = True  # Get rid of the javascript and the style
    cleaner.style           = True

    cleaner.meta            = False # Keeping the meta tags is important for page redirection purposes
    cleaner.safe_attrs_only = False

    return cleaner.clean_html(html)

def parse_academic_year(html):
    """Searches the html for r"(\\d\\d\\d\\d)/\\d\\d\\d\\d" and returns the
    first occurrence as an int. If no match is found, it will raise
    a AttributeError exception
    Eg: parse_academic_year("2018/2019 2020/2021") -> 2018
    """
    html = str(html) # must be a string (bs4 objects also work)

    academic_year = int(re.findall(r"(\d\d\d\d)/\d\d\d\d", html)[0])
    return academic_year

def get_image(url, params = None):
    """Fetches the image from the url and returns it as a PIL.Image object.
    If you need to be logged in to access the image you may want to check
    out the Credentials get_image function"""
    request = requests.get(url, params)
    image = Image.open(io.BytesIO(request.content))

    return image




import re as _re
import urllib as _urllib

import bs4 as _bs4
import PIL as _PIL

from . import _internal_utils as _utils
from . import cache as _cache

__all__ = ["Teacher"]

class Teacher:
    """This class represents a FEUP teacher as seen from their sigarra webpage.

    Note:
        Some attributes may be set to None, depending on whether or not
        that attribute was able to be parsed. For example: personal_webpage, email, and
        presentation are not always available in teacher pages.

    Args:
        p_codigo (int): The id of the teacher
        use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra
        base_url (:obj:`str`, optional): The url of the faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")


    Attributes:
        p_codigo         (int): The id of the teacher
        name             (str): The name of the teacher
        acronym          (str): The acronym of the teacher
        status           (str): The status of the teacher
        links            (tuple(str)): Urls from the teacher page (including :any:`Teacher.personal_webpage`, if present)
        personal_webpage (str): Url of the teacher's personal page
        url              (str): Url of the teacher's sigarra page
        voip             (int)
        email            (str)
        rooms            (str)
        category         (str)
        career           (str)
        profession       (str)
        department       (str)
        presentation     (str): The presentation of this teacher
        base_url         (str): The url of the teacher's faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")

    Example::

        from feupy import Teacher

        jlopes = Teacher(230756)

        print(jlopes.name)
        # João António Correia Lopes

        print(jlopes.acronym)
        # JCL

        print(jlopes.personal_webpage)
        # http://www.fe.up.pt/~jlopes/
    """
    __slots__ = ["p_codigo", "name", "acronym", "status", "links", "personal_webpage", "url", "voip",
                 "email", "rooms", "category", "career", "profession", "department", "presentation", "base_url"]

    def __init__(self, p_codigo : int, use_cache : bool = True, base_url : str = "https://sigarra.up.pt/feup/en/"):

        for attribute in self.__slots__:
            setattr(self, attribute, None)

        self.p_codigo = p_codigo
        self.base_url = base_url
        self.url = self.base_url + _utils.SIG_URLS["teacher"] + "?" + _urllib.parse.urlencode({"p_codigo" : str(p_codigo)})
        
        html = _cache.get_html(url = self.url, use_cache = use_cache) # Getting the html
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "O funcionário indicado não foi encontrado." in html:
            # The teacher's page is not here
            # Maybe it could be in another faculty?
            try:
                html = _cache.get_html(url = self.base_url + _utils.SIG_URLS["redirection page"] + "?" + _urllib.parse.urlencode({"pct_codigo" : str(p_codigo)}), use_cache = use_cache)
                soup = _bs4.BeautifulSoup(html, "lxml")
                self.url = soup.find("a")["href"]
                #########################
                if "/pt/" in self.url:
                    self.url = self.url.replace("/pt/", "/en/") # I want the page in english
                else:
                    index = self.url.index(_utils.SIG_URLS["teacher"])
                    self.url = self.url[:index] + "en/" + self.url[index:]
                #########################
                index = self.url.index(_utils.SIG_URLS["teacher"])
                self.base_url = self.url[:index]

                html = _cache.get_html(url = self.url, use_cache = use_cache) # Getting the html
                soup = _bs4.BeautifulSoup(html, "lxml")
            except:
                raise ValueError(f"Teacher with p_codigo {p_codigo} doesn't exist")

        personal_info = soup.find("div", {"class" : "informacao-pessoal-dados-dados"})

        for row in personal_info.find_all("tr"):
            if "Name:" in str(row):
                self.name = row.find_all("td")[1].b.string

                if row.find("a") != None:
                    self.personal_webpage = row.a["href"]
            
            elif "Acronym:" in str(row) or "Sigla:" in str(row):
                self.acronym = row.find_all("td")[1].string
            
            elif "Status:" in str(row) or "Estado:" in str(row):
                self.status = row.find_all("td")[1].string

            elif "Institutional E-mail:" in str(row) or "Email" in str(row):
                self.email = row.a.contents[0] + "@" + row.a.contents[-1]
                
            elif "Voip:" in str(row):
                self.voip = int(row.find_all("td")[1].string)

            elif "Rooms:" in str(row) or "Salas:" in str(row):
                self.rooms = row.find_all("td")[1].a.string

        self.links = tuple(tag["href"] for tag in soup.find("table", {"class" : "tabelasz"}).find_all("a"))

        functions_div = soup.find("div", {"class" : "informacao-pessoal-funcoes"})

        for td in functions_div.find_all("td", {"class" : "topo"}):

            if "Department:" in str(td):

                for row in td.find_all("tr"):
                    if "Category:" in str(row) or "Categoria:" in str(row):
                        self.category = row.find_all("td")[1].string
                    
                    elif "Career:" in str(row) or "Carreira:" in str(row):
                        self.career = row.find_all("td")[1].string

                    elif "Professional Group:" in str(row):
                        self.profession = row.find_all("td")[1].string
                    
                    elif "Department:" in str(row):
                        self.department = row.find_all("td")[1].string.strip()
                
                break
            
        
        personal_presentation = soup.find("div", {"class" : "informacao-pessoal-apresentacao"})

        if personal_presentation != None:
            self.presentation = personal_presentation.text.strip()
    
    def picture(self) -> _PIL.Image.Image:
        """Returns a picture of the teacher as a :obj:`PIL.Image.Image` object.
        
        Returns:
            A :obj:`PIL.Image.Image` object
        """
        return _utils.get_image(self.base_url + _utils.SIG_URLS["picture"], {"pct_cod" : str(self.p_codigo)})

    @classmethod
    def from_url(cls, url : str, use_cache : bool = True, base_url : str = "https://sigarra.up.pt/feup/en/"):
        """Scrapes the teacher webpage from the given url and returns a :obj:`Teacher` object.

        Args:
            url (str): The url of the teacher's sigarra page
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra
            base_url (:obj:`str`, optional): The url of the faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")
        
        Returns:
            A :obj:`Teacher` object

        Example::
        
            from feupy import Teacher

            url = "https://sigarra.up.pt/feup/en/func_geral.formview?p_codigo=230756"
            jlopes = Teacher.from_url(url)

            print(jlopes.name)
            # João António Correia Lopes
        """        
        matches = _re.findall(r"p_codigo=(\d+)$", url)
        
        if len(matches) == 0:
            raise ValueError(f"from_url() 'url' argument \"{url}\" is not a valid teacher url")
        
        p_codigo = int(matches[0])

        matches = _re.findall(r"^https?://sigarra\.up\.pt/(\w+)/", url)
        if len(matches) == 1:
            base_url = f"https://sigarra.up.pt/{matches[0]}/en/"

        return Teacher(p_codigo, use_cache, base_url = base_url)
    
    @classmethod
    def from_a_tag(cls, bs4_tag : _bs4.Tag, use_cache : bool = True, base_url : str = "https://sigarra.up.pt/feup/en/"):
        """Scrapes the teacher webpage from the given :obj:`bs4.tag` object and returns a :obj:`Teacher` object.
        
        Args:
            bs4_tag (:obj:`bs4.tag`):
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra
            base_url (:obj:`str`, optional): The url of the faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")
        
        Returns:
            A :obj:`Teacher` object
        """

        if bs4_tag.name != "a":
            raise ValueError(f"from_a_tag() 'bs4_tag' argument must be an anchor tag, not '{bs4_tag.name}'")
        
        return Teacher.from_url(bs4_tag["href"], use_cache, base_url = base_url)
    
    
    # Comparisons between teachers are made with the p_codigo
    def __eq__(self, other):
        if isinstance(other, Teacher):
            return self.p_codigo == other.p_codigo
        else:
            return NotImplementedError
        
    def __gt__(self, other):
        if isinstance(other, Teacher):
            return self.p_codigo > other.p_codigo
        else:
            return NotImplementedError
    
    def __ge__(self, other):
        if isinstance(other, Teacher):
            return self.p_codigo >= other.p_codigo
        else:
            return NotImplementedError
    
    def __lt__(self, other):
        if isinstance(other, Teacher):
            return self.p_codigo < other.p_codigo
        else:
            return NotImplementedError
    
    def __le__(self, other):
        if isinstance(other, Teacher):
            return self.p_codigo <= other.p_codigo
        else:
            return NotImplementedError
    
    @property
    def __dict__(self): # This is done for compatibility reasons (vars)
        return {attribute : getattr(self, attribute) for attribute in self.__slots__}

    def __hash__(self):
        return hash(self.p_codigo)
    
    def __repr__(self):
        if self.base_url == "https://sigarra.up.pt/feup/en/":
            return f"Teacher({self.p_codigo})"
        else:
            return f"Teacher({self.p_codigo}, base_url = {self.base_url})"
    
    def __str__(self):
        return f"{self.name} ({self.acronym})"



import getpass as _getpass
import os as _os
import re as _re
import urllib as _urllib

import requests as _requests
import requests_futures.sessions as _sessions

from . import _internal_utils as _utils

__all__ = ["Credentials"]

class Credentials:
    """This class represents your login credentials. It also
    features a non-persistent cache that stores htmls of urls
    previously accessed.

    Args:
        username (:obj:`int` or :obj:`str`, optional): Your username.
            You will be prompted for your username if you don't pass a username as an argument
        password (:obj:`str`, optional): Your password.
            You will be prompted for your password if you don't pass a password as an argument
        base_url (:obj:`str`, optional): The url of the faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")
    
    Attributes:
        username (int or str): Your username (e.g. 201806185)
        session  (:obj:`requests.Session`): A :obj:`requests.Session` object that holds the login cookies
        cache    (dict): A dictionary that maps a url string to an html string
        base_url (str): The url of your faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")

    Example::

        from feupy import Credentials

        username = 201806185
        password = "https://www.youtube.com/watch?v=dQw4w9WgXcQ"

        creds = Credentials(username, password)

        # You may instead enter your username and password at runtime
        creds = Credentials()
        # Username?
        # :> up201806185
        # Password for 201806185?
        # :>
    """
    __slots__ = ["username", "session", "cache", "base_url"]

    def __init__(self, username = None, password : str = None, base_url : str = "https://sigarra.up.pt/feup/en/"):
        #This is a modified version of the https://github.com/msramalho/sigpy login function

        AUTH_FAIL = "O conjunto utilizador/senha não é válido." # If this string appears on the HTML answer, the login creds aren't valid:(

        if username == None:
            username = input("Username?\n:> ")
        if password == None:
            password = _getpass.getpass(f"Password for {username}?\n:> ")
        
        session = _requests.Session()
        credencials = {'p_user': username, 'p_pass': password}
        request = session.post(base_url + _utils.SIG_URLS["authentication"], params = credencials) # Sending the creds to sigarra
        if AUTH_FAIL in request.text:
            raise ValueError("The login credencials aren't valid")
        
        self.username = username
        self.session  = session
        self.cache    = {}
        self.base_url = base_url
    
    def get_html(self, url : str, params : dict = {}) -> str:
        """Functionally equivalent to ``requests.get(url, params).text``,
        with scripts and styles removed. If the result is already in cache,
        the method will just return the value from the cache instead of
        making a web request.

        Args:
            url (str): The url of the html to be fetched
            params (:obj:`dict`, optional): the query portion of the url, should you want to include a query
        
        Returns:
            A string which is the html from the requested page url
        """
        if params != {}:
            url = url + "?" + _urllib.parse.urlencode(params, doseq = True) # Emulating the RequestsIII library
        
        if url in self.cache:
            pass
        else: # We need to get the page from the web
            request = self.session.get(url) # Getting the webpage
            request.raise_for_status()  # If the request fails, I want to know about it
            html = request.text

            self.cache[url] = _utils.trim_html(html)
        
        return self.cache[url]

    def get_html_async(self, urls, n_workers : int = 10):
        """:func:`Credentials.get_html`, but async, give or take.

        Takes a list (or any iterable) of urls and returns a corresponding generator of htmls.
        The htmls have their scripts and styles removed and are stored in cache.

        Args:
            urls (iterable(str)): The urls to be accessed
            n_workers (:obj:`int`, optional): The number of workers.
        
        Returns:
            An str generator
        """
        
        urls = tuple(urls)

        work_queue = (url for url in urls if url not in self.cache)

        with _sessions.FuturesSession(max_workers = n_workers) as session:
            session.cookies.update(self.session.cookies) # Getting the cookies from the object session to this assynchronous session

            futures = [session.get(url) for url in work_queue]

            for future in futures:
                request = future.result()

                if request.status_code != 200:
                    continue # GTFO
                
                url  = request.url
                html = request.text

                self.cache[url] = _utils.trim_html(html)
        
        return (self.get_html(url) for url in urls)
    
    def download(self, url : str, folder_path : str) -> str:
        """Downloads the file from the given url and saves it to the given folder path.
        If the path doesn't exist, it will be created automatically.
        
        Args:
            url (str): The url of the file to be downloaded
            folder_path (str): The path of the folder you want to download the file to
        
        Returns:
            The path of the file as an str

        """

        # https://stackoverflow.com/questions/16694907/download-large-file-in-python-with-requests
        # https://stackoverflow.com/questions/31804799/how-to-get-pdf-filename-with-python-requests

        if "conteudos_service.conteudos_cont" not in url:
            raise ValueError(f"the url argument {url} is not a valid sigarra download url")
        
        # create the cache folder if it does not exist
        if not _os.path.exists(folder_path):
            _os.makedirs(folder_path)
        
        with self.session.get(url) as r:
            r.raise_for_status()
            filename = _re.findall("filename=(.+)", r.headers['content-disposition'])[0][1:-1]
            
            path = _os.path.join(folder_path, filename)
            with open(path, 'wb') as f:
                for chunk in r.iter_content(chunk_size= 2 ** 15):
                    if chunk: # filter out keep-alive new chunks
                        f.write(chunk)
        
        return path
    
    @property
    def __dict__(self): # This is done for compatibility reasons (vars)
        return {attribute : getattr(self, attribute) for attribute in self.__slots__}
    
    def __repr__(self):
        if self.base_url == "https://sigarra.up.pt/feup/en/":
            return f"Credentials({self.username})"
        else:
            return f"Credentials({self.username}, base_url = {self.base_url})"



import datetime as _datetime
import re as _re
import urllib as _urllib

import bs4 as _bs4

from . import _Credentials
from . import cache as _cache
from . import exams as _exams
from . import _Student
from . import _Teacher
from . import timetable as _timetable
from . import _internal_utils as _utils

__all__ = ["CurricularUnit"]

class CurricularUnit:
    """This class represents a FEUP curricular unit.

    Args:
        pv_ocorrencia_id (int): The id of the curricular unit
        use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra
        base_url (:obj:`str`, optional): The url of the faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")

    Attributes:
        pv_ocorrencia_id   (int): The id of the curricular unit
        url                (str): The url of the curricular unit page
        name               (str): The name of the curricular unit
        code               (str)
        acronym            (str): The acronym of the curricular unit
        academic_year      (int): The academic year in which this curricular unit was taught
        semester           (str): The semester in which this curricular unit was taught (either '1', '2', or 'A')
        has_moodle         (bool): Whether or not this curricular unit has a moodle page
        is_active          (bool): Whether or not this curricular unit is active
        webpage_url        (str or None): The webpage of this curricular unit, if it exists. Otherwise it's set to None
        number_of_students (int)
        curricular_years   (tuple(int)): usually 1-5
        ECTS_credits       (float)
        regents            (tuple(:obj:`Teacher`))
        teachers           (tuple(:obj:`Teacher`))
        text               (string): Basically a text dump starting from "Teaching language"
        base_url           (str): The url of the curricular unit's faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")

    Example::

        from feupy import CurricularUnit
        from pprint import pprint

        mpcp = CurricularUnit(419989)

        print(mpcp.name)
        # Microprocessors and Personal Computers

        print(mpcp.acronym)
        # MPCP

        print(mpcp.semester)
        # 2

        for teacher in mpcp.teachers:
            print(teacher.name)
        # João Paulo de Castro Canas Ferreira
        # Bruno Miguel Carvalhido Lima
        # António José Duarte Araújo
        # João Paulo de Castro Canas Ferreira
        # João Paulo Filipe de Sousa

        pprint(vars(mpcp))
        # You'll get something like this:
        {'ECTS_credits': 6.0,
        'academic_year': 2018,
        'acronym': 'MPCP',
        'code': 'EIC0016',
        'curricular_years': (1,),
        'has_moodle': True,
        'is_active': True,
        'name': 'Microprocessors and Personal Computers',
        'number_of_students': 220,
        'pv_ocorrencia_id': 419989,
        'regents': (Teacher(210963),),
        'semester': '2',
        'teachers': (Teacher(210963),
                    Teacher(547486),
                    Teacher(211636),
                    Teacher(210963),
                    Teacher(210660)),
        'text': 'Teaching language\\n'
                'Portuguese\\n'
                'Objectives\\n'
                'BACKGROUND\\n'
                'The PC-compatible desktop and mobile platforms are an everyday tool '
                'in modern societies. Their architecture reflects the current '
                'technological development, but also defines the limits of the '
                "computer's capabilities and performance. Variants of the ARM "
                'instruction set are used today n most mobile platforms (tablets, '
                'mobile phones)in use today. Both system architecture and ISA have a '
                'deep impact on the day-to-day practice of informatics engineers.\\n'
                'SPECIFIC AIMS\\n'
                'The ... etc',
        'url': 'https://sigarra.up.pt/feup/en/ucurr_geral.ficha_uc_view?pv_ocorrencia_id=419989',
        'webpage_url': None}
    """
    __slots__ = ["pv_ocorrencia_id", "url", "name", "code", "acronym", "academic_year", "semester", "has_moodle", "is_active", "webpage_url", 
                 "number_of_students", "curricular_years", "ECTS_credits", "regents", "teachers", "text", "base_url"]
    
    def __init__(self, pv_ocorrencia_id : int, use_cache : bool = True, base_url : str = "https://sigarra.up.pt/feup/en/"):

        self.pv_ocorrencia_id = pv_ocorrencia_id
        self.base_url = base_url
        self.url = self.base_url + _utils.SIG_URLS["curricular unit"] + "?" + _urllib.parse.urlencode({"pv_ocorrencia_id" : str(pv_ocorrencia_id)})
        
        html = _cache.get_html(url = self.url, use_cache = use_cache) # Getting the html
        soup = _bs4.BeautifulSoup(html, "lxml")

        if soup.find("meta", {"http-equiv" : "Refresh"}) != None:
            self.url = soup.find("a")["href"]

            if "/pt/" in self.url:
                self.url = self.url.replace("/pt/", "/en/") # I want the page in english
            else:
                index = self.url.index("ucurr_geral.ficha_uc_view")
                self.url = self.url[:index] + "en/" + self.url[index:]

            html = _cache.get_html(url = self.url, use_cache = use_cache) # Getting the html
            soup = _bs4.BeautifulSoup(html, "lxml")
        
        index = self.url.index(_utils.SIG_URLS["curricular unit"])
        self.base_url = self.url[:index]
        
        if "The School responsible for the occurrence was not found." in html or "Não foi encontrada a ocorrência especificada." in html:
            raise ValueError(f"Curricular unit with pv_ocorrencia_id {pv_ocorrencia_id} doesn't exist")
       
        contents = soup.find("div", {"id" : "conteudoinner"})

        self.name = contents.find_all("h1")[1].string

        first_table = contents.find("table")
        self._parse_first_table(first_table)

        self.academic_year = _utils.parse_academic_year(contents)

        self._parse_matches(contents)

        self.has_moodle = "UC tem página no Moodle" in str(contents)

        second_table = first_table.find_next("table", {"class" : "formulario"})
        self._parse_second_table(second_table)
        
        third_table = second_table.find_next("table", {"class" : "dados"})
        self._parse_third_table(third_table)
        
        self._parse_teachers(contents)

        text = contents.text
        beggining_index = text.find("Teaching language")
        self.text = text[beggining_index:]


    def _parse_first_table(self, first_table) -> None:
        self.code    = _utils.scrape_html_table(first_table)[0][1]
        self.acronym = _utils.scrape_html_table(first_table)[0][4]


    def _parse_matches(self, contents):
        matches = _re.findall(r"Instance: \d\d\d\d/\d\d\d\d - (\d)S", str(contents))
        if len(matches) == 1:
            self.semester = matches[0]
        else:
            matches = _re.findall(r"Instance: \d\d\d\d/\d\d\d\d - A", str(contents))
            
            if len(matches) == 1:
                self.semester = 'A'
            else:
                self.semester = None


    def _parse_second_table(self, second_table) -> None:
        self.webpage_url = None
        for label, info in _utils.scrape_html_table(second_table):
            if   label == "Active? ":           # Usually the first row on the table
                self.is_active = "Yes" in info
            elif label == "Web Page: ":         # Usually the second row on the table (optional)
                self.webpage_url = info


    def _parse_third_table(self, third_table) -> None:
        rows = third_table.find_all("tr", {"class" : "d"})
        table = [row.find_all("td") for row in rows]
        
        # See https://sigarra.up.pt/feup/en/ucurr_geral.ficha_uc_view?pv_ocorrencia_id=436840
        # for the reason why if "len(row) >= 2" is here
        n_students = (int(row[1].string) for row in table if len(row) >= 2) 
        self.number_of_students = sum(n_students)

        if int(table[0][0]["rowspan"]) == 1: # Only 1 curricular year
            self.curricular_years = (int(table[0][3].string),) # A tuple
        else:
            self.curricular_years = [int(table[0][3].string)]
            for row in table[1:]:
                self.curricular_years.append(int(row[0].string))
            
            self.curricular_years = tuple(self.curricular_years)

        self.ECTS_credits    = float(table[0][5].string.replace(',', '.'))


    def _parse_teachers(self, contents):
        teachers_div   = contents.find("div", {"class" : "horas"})
        if teachers_div == None:
            teachers_links = filter(lambda tag: "func_geral.formview" in str(tag), contents.find_all("a"))
        else:
            teachers_links = filter(lambda tag: "func_geral.formview" in tag["href"], teachers_div.find_all("a"))

        teachers_urls = [self.base_url + tag["href"] for tag in teachers_links]
        _cache.get_html_async(teachers_urls)

        self.teachers  = tuple(_Teacher.Teacher.from_url(url, base_url = self.base_url) for url in teachers_urls)

        regents_div    = contents.find("div", {"class" : "responsabilidades"})
        if regents_div == None:
            self.regents = ()
        else:
            regents_links  = filter(lambda tag: "p_codigo" in tag["href"], regents_div.find_all("a")) # If "p_codigo" is in the url, then it's a teacher
            self.regents   = tuple((_Teacher.Teacher.from_a_tag(link, base_url = self.base_url) for link in regents_links))


    def contents(self, credentials : _Credentials.Credentials) -> dict:
        """Returns a nested dictionary structure, where each dictionary represents a folder.
        Every dictionary maps a string (the folder or file name) to either a dictionary
        (a nested folder) or a tuple (a file or a link).

        A tuple representing a file is made of 4 attributes:
            0. file type, either "file" or "link" (str)
            1. url (Note: you can download files by passing this url to :func:`Credentials.download`) (str)
            2. info (str or None)
            3. upload date (Note: links will have this attribute always set to today) (:obj:`Datetime.Date`)

        Args:
            credentials (:obj:`Credentials`): A :obj:`Credentials` object

        Returns:
            A nested dictionary structure

        Example::
        
            from feupy import CurricularUnit, Credentials
            from pprint import pprint

            mpcp = CurricularUnit(419989)

            creds = Credentials()

            pprint(mpcp.contents(creds))
            # You'll get something like this:
            {'A folder': {'folderception': {'A file': ('file',
                                                    'https://sigarra.up.pt/feup/pt/conteudos_service.conteudos_cont?pct_id=012345&pv_cod=06ahastCa',
                                                    None,
                                                    datetime.date(2018, 10, 2)),
                                            'Python': ('file',
                                                    'https://sigarra.up.pt/feup/pt/conteudos_service.conteudos_cont?pct_id=012346&pv_cod=06TtaSaPH7',
                                                    'aaaaaa',
                                                    datetime.date(2018, 11, 13)),
                                            'Simulator': ('link',
                                                        'https://github.com/hneemann/Digital',
                                                        'It\\'s a simulator.',
                                                        datetime.date(2019, 5, 25))},
                        'jhbh': {'aa': ('link',
                                        'https://salmanarif.bitbucket.io/visual/index.html',
                                        'aaaaaa',
                                        datetime.date(2019, 8, 2)),
                                'bb': ('file',
                                        'https://sigarra.up.pt/feup/pt/conteudos_service.conteudos_cont?pct_id=012347&pv_cod=06WjyvGato',
                                        'banana',
                                        datetime.date(2018, 10, 12))}}}
        """
        html = credentials.get_html(self.url) # curricular unit page
        soup = _bs4.BeautifulSoup(html, "lxml")

        tag = soup.find("a", {"title" : "Contacts"})

        if tag == None:
            return {}
            # The curricular unit either doesn't have a contents page
            # or you don't have access to it
            # I can't distinguish between these two situations, therefore I'm
            # returning an empty dict. I prefer this approach to raising an (perhaps erroneous) exception
        
        url = self.base_url.replace("/en/", "/pt/") + tag["href"]

        html = credentials.get_html(url) # contents page
        soup = _bs4.BeautifulSoup(html, "lxml")
        content = soup.find("div", {"id" : "conteudoinner"})

        if "Não existem conteúdos para ver" in html:
            return {}
        
        def has_pct_grupo(a):
            try:
                int(a["name"])
                return True
            except:
                return False

        pct_grupos = (a["name"] for a in content.find_all("a") if has_pct_grupo(a))
        query = "&".join(f"pct_grupo={grupo}" for grupo in pct_grupos)
        url = url + "&" + query 

        html = credentials.get_html(url) # contents page with all the folders open
        soup = _bs4.BeautifulSoup(html, "lxml")
        content = soup.find("div", {"id" : "conteudoinner"})
        
        files_and_dirs = [p for p in content.find_all("p") if p.has_attr("class")]
        
        def convert_list_to_dict(files_and_dirs_list : list, result : dict, conteudosnivel : int = 1):
            """Please use protective glasses when looking at this function"""
            for i, p in enumerate(files_and_dirs_list):
                if p["class"] == [f"conteudosnivel{conteudosnivel}"]:
                    if "/feup/pt/imagens/Pasta" in str(p): # It's a directory
                        name = p.find_all("a")[-1].string
                        result[name] = {}

                        for i1, p1 in enumerate(files_and_dirs_list):
                            if i1 <= i:
                                continue

                            if p1["class"] == [f"conteudosnivel{conteudosnivel}"]:
                                convert_list_to_dict(files_and_dirs_list[i + 1: i1], result[name], conteudosnivel + 1)
                                break
                        else:
                            convert_list_to_dict(files_and_dirs_list[i + 1:], result[name], conteudosnivel + 1)
                    
                    else: # It's a file (or a link)
                        if p.a == None: # If there's no link, just ignore it
                            continue

                        file_type = "link" if p.find("span", {"class" : "t"}) == None else "file"
                        name = p.a.string
                        try:
                            info = p.find("span", {"class" : "textopequenoconteudos"}).string
                        except:
                            info = None

                        if file_type == "file":
                            
                            upload_time = _datetime.date(*map(int, p.find("span", {"class" : "t"}).string.split("-")[1].split("/")))
                            url = self.base_url.replace("/en/", "/pt/") + p.a["href"]
                        else:
                            upload_time = _datetime.date.today() # A placeholder value
                            url = p.a["href"]

                        result[name] = (file_type, url, info, upload_time)

        result = {}
        convert_list_to_dict(files_and_dirs, result, 1)

        return result

    def students(self, credentials : _Credentials.Credentials, use_cache : bool = True) -> list:
        """Returns the students of this curricular unit as a list of tuples.
        
        Each tuple has 4 elements:
            0. student (:obj:`Student`)
            1. status (str)
            2. number of registrations (int)
            3. type of student (str)
        
        Args:
            credentials (:obj:`Credentials`): A :obj:`Credentials` object
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra

        Returns:
            A list of tuples

        Example::

            from feupy import CurricularUnit, Credentials
            from pprint import pprint

            mpcp = CurricularUnit(419989)

            creds = Credentials()

            pprint(mpcp.students(creds))
            # You'll get something like this:
            [
                ...,
                (Student(201812345), 'Ordinário', 1, 'Normal'),
                ...
            ]
        """

        data   = []
        tables = []

        def format_table(tags_list, index): # Note: what this function returns doesn't matter,
            if index == 0:                  # we're appending the information in the tables to data
                return None                               
            
            student_username = int(tags_list[0].string)
            student = _Student.Student(student_username, base_url = self.base_url)

            status = tags_list[2].string.strip()

            number_of_registrations = int(tags_list[3].string)

            student_type = tags_list[4].string

            data.append((student, status, number_of_registrations, student_type))
                    
        n_pages    = (self.number_of_students // 50) + 1
        pages_urls = []
        for n in range(1, n_pages + 1):
            url = self.base_url + _utils.SIG_URLS["curricular unit students"] + "?" + _urllib.parse.urlencode({"pv_ocorrencia_id" : str(self.pv_ocorrencia_id), "pv_num_pag" : str(n)})
            pages_urls.append(url)
        
        for html in credentials.get_html_async(pages_urls):
            soup = _bs4.BeautifulSoup(html, "lxml")
            table = soup.find("table", {"class" : "dadossz"})
            tables.append(table)


        student_urls = []
        for table in tables: # get all the student urls
            student_usernames = _re.findall(r"(\d\d\d\d\d\d\d\d\d)", str(table))
            student_urls.extend(self.base_url + _utils.SIG_URLS["student page"] + "?" + _urllib.parse.urlencode({"pv_num_unico" : username}) for username in student_usernames)
        _cache.get_html_async(student_urls, use_cache = use_cache) # Refreshing the cache

        for table in tables:
            _utils.scrape_html_table(table, format_table) # get the data in the tables
        
        return data
    
    def timetable(self, credentials : _Credentials.Credentials) -> list:
        """Returns the curricular unit's current timetable as a list of dictionaries if possible, otherwise returns None.
        (see :func:`timetable.parse_current_timetable` for more info)
        
        Returns:
            A list of dicts
        """
        html = self.base_url.replace("/en/", "/pt/") + credentials.get_html(_utils.SIG_URLS["curricular unit timetable"], {"pv_ocorrencia_id" : self.pv_ocorrencia_id})
        soup = _bs4.BeautifulSoup(html, "lxml")

        return _timetable.parse_current_timetable(credentials, soup.a["href"])
    
    def all_timetables(self, credentials : _Credentials.Credentials) -> dict:
        """Parses all the timetables related to this curricular unit
        (see :obj:`timetable.parse_timetables` for further info).
        
        Returns:
            A dictionary which maps a tuple with two :obj:`datetime.date` objects,
            start and finish (the time span in which this timetable is valid), to a
            list of dictionaries (see :obj:`timetable.parse_timetable` for an example
            of such a list).
        """
        html = self.base_url.replace("/en/", "/pt/") + credentials.get_html(_utils.SIG_URLS["curricular unit timetable"], {"pv_ocorrencia_id" : self.pv_ocorrencia_id})
        soup = _bs4.BeautifulSoup(html, "lxml")

        return _timetable.parse_timetables(credentials, soup.a["href"])

    def other_occurrences(self, use_cache : bool = True) -> tuple:
        """Returns the occurrences of this curricular unit from other years as a
        tuple of :obj:`CurricularUnit` objects.

        Args:
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra

        Returns:
            A tuple of :obj:`CurricularUnit` objects
        
        Example::

            from feupy import CurricularUnit
            from pprint import pprint

            mpcp = CurricularUnit(419989)

            pprint(mpcp.other_occurrences())
            # You'll get something like this:
            (CurricularUnit(436431),
            CurricularUnit(419989),
            CurricularUnit(399884),
            CurricularUnit(384929),
            CurricularUnit(368695),
            CurricularUnit(350451),
            CurricularUnit(333111),
            CurricularUnit(272647),
            CurricularUnit(272646),
            CurricularUnit(272645),
            CurricularUnit(272644),
            CurricularUnit(272642),
            CurricularUnit(272641),
            CurricularUnit(272640),
            CurricularUnit(272639))
        """
        html = _cache.get_html(self.url)
        soup = _bs4.BeautifulSoup(html, "lxml")

        tag = soup.find("a", {"title" : "Other occurrences"})
        url = self.base_url + tag["href"]

        html = _cache.get_html(url)         # Now we have the html of the "other occurrences" page 
        soup = _bs4.BeautifulSoup(html, "lxml")

        table = soup.find_all("table", {"class" : "dados"})[1]
        course_tags = table.find_all("a")

        _cache.get_html_async((self.base_url + a_tag["href"] for a_tag in course_tags), use_cache = use_cache) # Refresh the cache

        return tuple(CurricularUnit.from_a_tag(a_tag, base_url = self.base_url) for a_tag in course_tags)
        
    def stats(self, credentials : _Credentials.Credentials) -> tuple:
        """Returns a tuple with 3 ints:
            0. number of registered students
            1. number of evaluated students
            2. number of approved students

        Args:
            credentials (:obj:`Credentials`): A :obj:`Credentials` object

        Returns:
            A tuple with 3 ints

        Example::

            from feupy import CurricularUnit, Credentials

            mpcp = CurricularUnit(419989)

            creds = Credentials()

            print(mpcp.stats(creds))
            # You'll get something like this:
            (220, 185, 162)
        """
        
        html = credentials.get_html(self.base_url + _utils.SIG_URLS["curricular unit statistics"], params = {"pv_ocorrencia_id" : str(self.pv_ocorrencia_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "Não foram encontrados estudantes inscritos na ocorrência indicada." in html:
            raise LookupError(f"The statistics for {self.__repr__()} have not been released yet")

        table = soup.find_all("table")[-1] # get the last table of the page
        table_data = table.find_all("td", {"class" : "k n"})[0:3] # get the first three tds

        registered, evaluated, approved = map(lambda tag: int(tag.string), table_data)

        return (registered, evaluated, approved)
    
    def grades_distribution(self, credentials : _Credentials.Credentials) -> dict:
        """Returns the distribution of the grades as a dict.

        Args:
            credentials (:obj:`Credentials`): A :obj:`Credentials` object
        
        Returns:
            A dictionary that maps the grade to the number of students that got that grade.

        Example::
        
            from feupy import CurricularUnit, Credentials
            from pprint import pprint

            mpcp = CurricularUnit(419989)

            creds = Credentials()

            pprint(mpcp.grades_distribution(creds))
            # You'll get something like this:
            {'RFC': 9,
            'RFE': 17,
            'RFF': 18,
                5: 4,
                6: 5,
                7: 7,
                8: 6,
                9: 4,
                10: 12,
                11: 11,
                12: 13,
                13: 21,
                14: 4,
                15: 12,
                16: 7,
                17: 13,
                18: 8,
                19: 7,
                20: 1}
        """

        html = credentials.get_html(self.base_url + _utils.SIG_URLS["curricular unit grades distribution"], params = {"pv_ocorrencia_id" : str(self.pv_ocorrencia_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "Não foram encontrados estudantes avaliados na ocorrência indicada." in html:
            raise LookupError(f"The statistics for {self.__repr__()} have not been released yet")
        
        table = soup.find_all("table")[-2] # get the second last table of the page
        tbody = table.find("tbody")
        result = {}

        def format_table(tags_list, index): # Note: what this function returns doesn't matter,
            try:                            # We're updating the result dictionary
                grade = int(tags_list[0].string)
            except:
                grade = tags_list[0].string
            
            n_students = int(tags_list[2].string)

            result[grade] = n_students
        
        _utils.scrape_html_table(tbody, format_table)

        return result

    def statistics_history(self, credentials : _Credentials.Credentials) -> list:
        """Returns a list of tuples each representing an academic year.

        Contents of a tuple:
            0. academic year        (int)
            1. registered students  (int)
            2. evaluated students   (int)
            3. approved students    (int)
            4. evaluated students average grade       (float)
            5. evaluated students standard deviation  (float)
            6. approved students  average grade       (float)
            7. approved students  standard deviation  (float)

        Args:
            credentials (:obj:`Credentials`): A :obj:`Credentials` object
        
        Returns:
            A list of tuples

        Example::

            from feupy import CurricularUnit, Credentials
            from pprint import pprint

            mpcp = CurricularUnit(419989)

            creds = Credentials()

            pprint(mpcp.statistics_history(creds))

            # You will get something like this:
            [(2016, 208, 152, 120, 12.43, 3.75, 13.4,  2.5),
             (2017, 203, 149, 113, 12.08, 3.08, 12.27, 3.0),
             (2018, 203, 170, 133, 12.51, 3.76, 13.53, 2.76),
             ... ]
        """
        
        html = credentials.get_html(self.base_url + _utils.SIG_URLS["curricular unit stats history"], params = {"pv_ocorrencia_id" : str(self.pv_ocorrencia_id), "pv_n_prev_alet" : "20"})
        soup = _bs4.BeautifulSoup(html, "lxml")

        table = soup.find_all("table")[-1] # get the last table of the page
        tbody = table.find("tbody")
        rows = [row for row in tbody.find_all("tr") if row["class"] == ['p']]

        result = []
        
        for row in rows:
            academic_year_td = row.find("td", {"class" : "k"})
            academic_year = int(academic_year_td.text[0:4])

            row_data = row.find_all("td", {"class" : "n"})

            registered, evaluated, approved = map(lambda tag: int(tag.string), row_data[0:3])

            evaluated_avg, evaluated_std_dev, approved_avg, approved_std_dev = map(lambda tag: float(tag.string), row_data[3:7])

            result.append((academic_year, registered, evaluated, approved, evaluated_avg, evaluated_std_dev, approved_avg, approved_std_dev))

        return result
    
    def classes(self, credentials : _Credentials.Credentials, use_cache : bool = True) -> dict:
        """Returns a dictionary which maps a class name to a list of students
        (the students of that class).
        
        Args:
            credentials (:obj:`Credentials`): A :obj:`Credentials` object
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra

        Returns:
            A dict

        Example::

            from feupy import CurricularUnit, Credentials
            from pprint import pprint

            mpcp = CurricularUnit(419989)

            creds = Credentials()

            pprint(mpcp.classes(creds))
            # You'll get something like this:
            {
                '1MIEIC01': [Student(201800000),
                             Student(201800001),
                             Student(201800002),
                             Student(201800003)],
                '1MIEIC02': [Student(201800004),
                             Student(201800005),
                             Student(201800006),
                             Student(201800007),
                             Student(201800008),
                             Student(201800009),
                             Student(201800010),
                             Student(201800011),
                             Student(201800012)],
                '...'     : [...]
            }
        """
        html = credentials.get_html(self.base_url + _utils.SIG_URLS["curricular unit classes"], params = {"pv_ocorrencia_id" : str(self.pv_ocorrencia_id)}) # this html redirects us to the url we want
        soup = _bs4.BeautifulSoup(html, "lxml")

        tags = soup.find_all("a")

        if len(tags) == 1:
            urls = [tags[0]["href"]]
        else:
            urls = [self.base_url + tag["href"] for tag in tags if "it_listagem.lista_turma_disciplina" in str(tag)]

        result = {}

        for url in urls:
            html = credentials.get_html(url)
            soup = _bs4.BeautifulSoup(html, "lxml")
            contents = soup.find("div", {"id" : "conteudo"})

            all_students_urls = (self.base_url + tag["href"] for tag in contents.find_all("a") if tag.parent.name == "td")
            _cache.get_html_async(all_students_urls, use_cache = use_cache) # refresh the cache
            
            title = contents.find("h3") # starting point

            while title.find_next("h3") != None:
                title = title.find_next("h3") # Skip to the next class

                class_name = _re.findall(r"Turma: ([^\s]+)", title.text)[0]

                table = title.find_next("table")

                if table == None:
                    students = []
                else:
                    students = [_Student.Student.from_a_tag(tag, base_url = self.base_url) for tag in table.find_all("a")]

                result[class_name] = students

        return result

    def exams(self, use_cache : bool = True):
        """Returns a list of the exams of this curricular unit.
        
        Args:
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra

        Returns:
            A list of dictionaries (see :func:`exams.exams` for more information about the dictionaries)
        """
        url = self.base_url.replace("/en/", "/pt/") + _utils.SIG_URLS["curricular unit exams"] + "?" + _urllib.parse.urlencode({"p_ocorr_id" : str(self.pv_ocorrencia_id)})

        return _exams.exams(url, use_cache, base_url = self.base_url)

    def results(self, credentials : _Credentials.Credentials, use_cache : bool = True) -> dict:
        """Returns a dictionary which maps a string representing the exams season
        ('Época Normal (2ºS)'/'Época Recurso (2ºS)') to a list of tuples.
        
        Each tuple in the list has two values:
            1. student (:obj:`Student`)
            2. grade (str or int)

        Args:
            credentials (:obj:`Credentials`): A :obj:`Credentials` object
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra

        Returns:
            A dict

        Example::
            
            from feupy import CurricularUnit, Credentials
            from pprint import pprint

            mpcp = CurricularUnit(419989)

            creds = Credentials()

            pprint(mpcp.results(creds))
            # You'll get something like this:
            {'Época Normal (2ºS)': [(Student(201800001), 10),
                                    (Student(201800002), 13),
                                    (Student(201800003), 10),
                                    (Student(201800004), 'RFE'),
                                    (Student(201800005), 'RFF'),
                                    (Student(201800006), 'RFF'),
                                    ...],
            'Época Recurso (2ºS)': [(Student(201800008), 11),
                                    (Student(201800009), 7),
                                    (Student(201800010), 8),
                                    (Student(201800011), 8),
                                    (Student(201800012), 'RFE'),
                                    (Student(201800013), 13),
                                    (Student(201800014), 5),
                                    (Student(201800019), 'RFC'),
                                    ...]}
        """
        html = credentials.get_html(self.base_url + _utils.SIG_URLS["curricular unit results"], params = {"pv_ocorr_id" : str(self.pv_ocorrencia_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "Não tem permissões para aceder a este conteúdo" in html:
            raise PermissionError("Your Credentials object does not have access to this curricular unit's results")
        
        tags = (tag for tag in soup.find_all("a") if "lres_geral.show_pauta_resul" in str(tag))

        result = {}

        for tag in tags:
            result[tag.string] = []

            url = self.base_url + tag["href"]

            html = credentials.get_html(url)
            soup = _bs4.BeautifulSoup(html, "lxml")

            table = soup.find("table", {"class" : "dadossz"})

            rows = table.find_all("tr", {"class" : "i"})

            student_urls = []
            for row in rows: # get all the student urls
                student_usernames = _re.findall(r"(\d\d\d\d\d\d\d\d\d)", str(row))
                student_urls.extend(self.base_url + _utils.SIG_URLS["student page"] + "?" + _urllib.parse.urlencode({"pv_num_unico" : username}) for username in student_usernames)
            _cache.get_html_async(student_urls, use_cache = use_cache) # Refreshing the cache

            for row in rows:
                username = int(row.find("td", {"class" : "l"}).string)

                grade_str = row.find("td", {"class" : "n"}).string
                try:
                    grade = int(grade_str)
                except:
                    grade = grade_str
                
                result[tag.string].append((_Student.Student(username, base_url = self.base_url), grade))
        
        return result

    @classmethod
    def from_url(cls, url : str, use_cache : bool = True, base_url : str = "https://sigarra.up.pt/feup/en/"):
        """Scrapes the curricular webpage from the given url and returns a :obj:`CurricularUnit` object.

        Args:
            url (str): The url of the curricular unit's sigarra page
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra
            base_url (:obj:`str`, optional): The url of the faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")
        
        Returns:
            A :obj:`CurricularUnit` object

        Example::
        
            from feupy import CurricularUnit

            url = "https://sigarra.up.pt/feup/pt/ucurr_geral.ficha_uc_view?pv_ocorrencia_id=419989"
            mpcp = CurricularUnit.from_url(url)

            print(mpcp.name)
            # Microprocessors and Personal Computers
        """

        matches = _re.findall(r"pv_ocorrencia_id=(\d+)$", url)
        
        if len(matches) == 0:
            raise ValueError(f"from_url() 'url' argument \"{url}\" is not a valid curricular unit url")
        
        pv_ocorrencia_id = int(matches[0])

        matches = _re.findall(r"^https?://sigarra\.up\.pt/(\w+)/", url)
        if len(matches) == 1:
            base_url = f"https://sigarra.up.pt/{matches[0]}/en/"

        return CurricularUnit(pv_ocorrencia_id, use_cache, base_url = base_url)
    
    @classmethod
    def from_a_tag(cls, bs4_tag : _bs4.Tag, use_cache : bool = True, base_url : str = "https://sigarra.up.pt/feup/en/"):
        """Scrapes the curricular unit webpage from the given :obj:`bs4.tag` object and returns a :obj:`CurricularUnit` object.
        
        Args:
            bs4_tag (:obj:`bs4.tag`):
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra
            base_url (:obj:`str`, optional): The url of the faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")
        
        Returns:
            A :obj:`CurricularUnit` object
        """

        if bs4_tag.name != "a":
            raise ValueError(f"from_a_tag() 'bs4_tag' argument must be an anchor tag, not '{bs4_tag.name}'")
        
        return CurricularUnit.from_url(bs4_tag["href"], use_cache, base_url = base_url)
    
    # Comparisons between curricular units are made with the pv_ocorrencia_id
    def __eq__(self, other):
        if isinstance(other, CurricularUnit):
            return self.pv_ocorrencia_id == other.pv_ocorrencia_id
        else:
            return NotImplementedError
        
    def __gt__(self, other):
        if isinstance(other, CurricularUnit):
            return self.pv_ocorrencia_id > other.pv_ocorrencia_id
        else:
            return NotImplementedError
    
    def __ge__(self, other):
        if isinstance(other, CurricularUnit):
            return self.pv_ocorrencia_id >= other.pv_ocorrencia_id
        else:
            return NotImplementedError
    
    def __lt__(self, other):
        if isinstance(other, CurricularUnit):
            return self.pv_ocorrencia_id < other.pv_ocorrencia_id
        else:
            return NotImplementedError
    
    def __le__(self, other):
        if isinstance(other, CurricularUnit):
            return self.pv_ocorrencia_id <= other.pv_ocorrencia_id
        else:
            return NotImplementedError
    
    @property
    def __dict__(self): # This is done for compatibility reasons (vars)
        return {attribute : getattr(self, attribute) for attribute in self.__slots__}

    def __hash__(self):
        return hash(self.pv_ocorrencia_id)
    
    def __repr__(self):
        if self.base_url == "https://sigarra.up.pt/feup/en/":
            return f"CurricularUnit({self.pv_ocorrencia_id})"
        else:
            return f"CurricularUnit({self.pv_ocorrencia_id}, base_url = {self.base_url})"
    
    def __str__(self):
        return f"{self.acronym} ({self.academic_year}/{self.academic_year + 1})" # e.g. ALGE (2019/2019)



from feupy import Credentials, cache
from pathlib import Path

from .FeupyTestCase import FeupyTestCase

print("In order to test methods that require "
      "priviledged access to sigarra, i.e. student access, "
      "a valid set set of credentials is required")

while True:
    try:
        creds = Credentials()
        break
    except ValueError:
        print("Invalid credentials, please try again")

# We need an empty cache
cache.cache = {}



import re as _re
import urllib as _urllib

import bs4 as _bs4

from . import _Credentials
from . import _internal_utils as _utils
from . import cache as _cache
from . import _CurricularUnit
from . import exams as _exams
from . import _Teacher

__all__ = ["Course"]

class Course:
    """This class represents a course as seen from its sigarra webpage.

    Args:
        pv_curso_id (int): The id of the course, for example MIEIC's id is 742
        pv_ano_lectivo (:obj:`int`, optional): The year of this course. It defaults to the current year (i.e. 2019, at the time of writing)
        use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra
        base_url (:obj:`str`, optional): The url of the faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")

    Attributes:
        pv_curso_id            (int): The id of the course
        pv_ano_lectivo         (int): The year of this course's page
        url                    (str): Url of the course's sigarra page
        name                   (str): The name the course
        official_code          (str): The official code
        directors              (tuple(:obj:`Teacher`)): The directors of this course
        acronym                (str): The acronym of this course
        text                   (str or None): The text that can be found in the course's page
        base_url               (str): The url of the course's faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")
        involved_organic_units (tuple(str)): All the faculties involved with this course

    Example::

        from feupy import Course

        mieic = Course(742)

        print(mieic.name)
        # Master in Informatics and Computing Engineering

        print(mieic.acronym)
        # MIEIC

        for director in mieic.directors:
            print(director.name)
        # João Carlos Pascoal Faria
        # Maria Cristina de Carvalho Alves Ribeiro
    """
    __slots__ = ["pv_curso_id","pv_ano_lectivo", "url", "name", "official_code", "directors", "acronym", "text", "base_url", "involved_organic_units"]

    def __init__(self, pv_curso_id : int, pv_ano_lectivo : int = None, use_cache : bool = True, base_url : str = "https://sigarra.up.pt/feup/en/"):
        
        if pv_ano_lectivo == None:
            pv_ano_lectivo = _utils.get_current_academic_year()

        self.pv_curso_id    = pv_curso_id
        self.base_url = base_url
        self.pv_ano_lectivo = pv_ano_lectivo

        payload = {
            "pv_curso_id"    : str(pv_curso_id),
            "pv_ano_lectivo" : str(pv_ano_lectivo)
        }
        self.url = self.base_url + _utils.SIG_URLS["course"] + "?" + _urllib.parse.urlencode(payload)
        
        html = _cache.get_html(url = self.url, use_cache = use_cache) # Getting the html
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "Course/Cycle of Studies nonexistent." in html or "The Organic Unit is not involved in teaching the course/CS." in html:
            raise ValueError(f"Course with pv_curso_id {pv_curso_id} and pv_ano_lectivo {pv_ano_lectivo} doesn't exist")
       
        contents = soup.find("div", {"id" : "conteudoinner"})

        self.name = contents.find_all("h1")[1].string

        info_box = contents.find("div", {"class" : "caixa-informativa"})

        directors_ids = _re.findall(r"vld_entidades_geral.entidade_pagina\?pct_codigo=(\d+)", html)
        self.directors  = tuple(_Teacher.Teacher(int(_id), base_url = self.base_url) for _id in directors_ids)

        for row in info_box.find_all("tr"):
            if "Acronym:" in str(row):
                self.acronym = row.find_all("td")[1].string
            
            elif "Official Code:" in str(row):
                self.official_code = row.find_all("td")[1].string
        
        try:
            self.text = contents.find("div", {"class" : "col-md-8 col-sm-6 col-xs-12"}).text
        except:
            self.text = None
        
        for h3 in soup.find_all("h3"):
            if h3.text == "Involved Organic Units":
                involved_organic_units = []

                div = h3.find_next("div")
                for a in div.find_all("a"):
                    # I'm well aware that hardcoding values is not good practice.
                    # However, trying to follow all the redirects has turned out to be rather unfeasable.
                    # Trust me, I tried and thoroughly failed:(
                    # If you think you can do better, PRs are always welcome:)

                    if "href" not in a.attrs:
                        continue

                    if   "letras.up.pt" in a["href"]:
                        involved_organic_units.append("https://sigarra.up.pt/flup/en/")
                    elif "fe.up.pt"     in a["href"]:
                        involved_organic_units.append("https://sigarra.up.pt/feup/en/")
                    elif "fep.up.pt"    in a["href"]:
                        involved_organic_units.append("https://sigarra.up.pt/fep/en/")
                    elif "fba.up.pt"    in a["href"]:
                        involved_organic_units.append("https://sigarra.up.pt/fbaup/en/")
                    elif "icbas.up.pt"  in a["href"]:
                        involved_organic_units.append("https://sigarra.up.pt/icbas/en/")
                    elif "fc.up.pt"     in a["href"]:
                        involved_organic_units.append("https://sigarra.up.pt/fcup/en/")
                    elif "fa.up.pt"     in a["href"]:
                        involved_organic_units.append("https://sigarra.up.pt/faup/en/")
                
                self.involved_organic_units = tuple(involved_organic_units)
                break
        else:
            self.involved_organic_units = (self.base_url,)

    def classes(self, credentials : _Credentials.Credentials) -> dict:
        """Returns a dictionary which maps the course's classes
        to their corresponding timetable links.

        Args:
            credentials (:obj:`Credentials`): A :obj:`Credentials` object
        
        Returns:
            A dictionary
        
        Example::

            from feupy import Course, Credentials
            from pprint import pprint

            mieic = Course(742)

            creds = Credentials()

            pprint(mieic.classes(creds))

            # You will get something like this:
            {'1MIEIC01':'https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=000001&pv_periodos=1&pv_ano_lectivo=2018',
            '1MIEIC02': 'https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=000002&pv_periodos=1&pv_ano_lectivo=2018',
            '1MIEIC03': 'https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=000003&pv_periodos=1&pv_ano_lectivo=2018',
            '1MIEIC04': 'https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=000004&pv_periodos=1&pv_ano_lectivo=2018',
            '1MIEIC05': 'https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=000005&pv_periodos=1&pv_ano_lectivo=2018',
            '1MIEIC06': 'https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=000006&pv_periodos=1&pv_ano_lectivo=2018',
            ...                                                                                                                   }
        
        You can use the functions provided by :doc:`timetable` to get the timetables from the urls.
        """

        payload = {
            "pv_curso_id"    : str(self.pv_curso_id),
            "pv_ano_lectivo" : str(self.pv_ano_lectivo),
            "pv_periodos"    : "1" # All the classes in that year
        }

        html = self.base_url.replace("/en/", "/pt/") + credentials.get_html(_utils.SIG_URLS["course classes"], payload)
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "Não existem dados para este ano letivo." in html:
            raise ValueError("No classes were available to be parsed")

        classes_tags = soup.find_all("a", {"class" : "t"})

        return {tag.string : self.base_url.replace("/en/", "/pt/") + tag["href"] for tag in classes_tags}


    def curricular_units(self, use_cache : bool = True) -> list:
        """Returns a list of :obj:`CurricularUnit` objects representing
        the curricular units of the course. Curricular units without
        a link are not included.
        
        Args:
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra

        Returns:
            A list of :obj:`CurricularUnit` objects


        Example::

            from feupy import Course
            from pprint import pprint

            mieic = Course(742)

            pprint(mieic.curricular_units())

            # You will get something like this:
            [CurricularUnit(419981),
             CurricularUnit(419982),
             CurricularUnit(419983),
             CurricularUnit(419984),
             CurricularUnit(419985),
             CurricularUnit(419986),
             CurricularUnit(419987),
             CurricularUnit(419988),
             CurricularUnit(419989),
             CurricularUnit(419990),
             CurricularUnit(419991),
             CurricularUnit(419992),
             CurricularUnit(419993),
             CurricularUnit(419994),
             CurricularUnit(419995),
             CurricularUnit(419996),
             ...                   ]
        """
        html = _cache.get_html(self.url, use_cache = use_cache)
        soup = _bs4.BeautifulSoup(html, "lxml")

        for tag in soup.find_all("a"):
            if "cur_geral.cur_planos_estudos_view" in str(tag):
                url = self.base_url + tag["href"]
                break
        else:
            raise ValueError("No study plan link was found on the course page")
        
        html = _cache.get_html(url, use_cache = use_cache)
        soup = _bs4.BeautifulSoup(html, "lxml")

        curricular_units_tags = (tag for tag in soup.find_all("a") if "ucurr_geral.ficha_uc_view" in str(tag))
        curricular_units_urls = [self.base_url + tag["href"] for tag in curricular_units_tags]

        # The following block of code assynchronously loads both the curricular units' and the teachers' webpages
        teachers_urls = []
        for uc_html in _cache.get_html_async(curricular_units_urls, use_cache = use_cache):
            uc_soup = _bs4.BeautifulSoup(uc_html, "lxml")

            for tag in uc_soup.find_all("a"):
                if "func_geral.formview" in str(tag):
                    teachers_urls.append(self.base_url + tag["href"])
        _cache.get_html_async(teachers_urls, use_cache = use_cache)

        res = set()
        for url in curricular_units_urls:
            try:
                res.add(_CurricularUnit.CurricularUnit.from_url(url, base_url = self.base_url))

            except ValueError as e:
                # It is possible that Sigarra can have dead links in the study plan
                if "Curricular unit with pv_ocorrencia_id" not in str(e):
                    raise e

        return list(res)

    
    def syllabus(self, use_cache : bool = True) -> list:
        """Returns a list of tuples containing more information about the curricular units of the course.
        
        Each tuple has 4 elements:
            0. keyword (str)
            1. course branch (either str or None, depending on whether or not the curricular unit belongs to a specific branch)
            2. curricular unit (:obj:`CurricularUnit`)
            3. is mandatory (bool)
        
        Args:
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra

        Returns:
            A list of tuples

        Example::

            from feupy import Course
            from pprint import pprint

            miem = Course(743)

            pprint(miem.syllabus())

            # You will get something like this:
            [
                ('Automation', 'Automation', CurricularUnit(418771), False),
                ('Automation', 'Automation', CurricularUnit(418770), False),
                ('Automation', 'Automation', CurricularUnit(418737), True),
                ('Automation', 'Automation', CurricularUnit(418746), True),
                ('Automation', 'Automation', CurricularUnit(418747), False),
                ('Automation', 'Automation', CurricularUnit(418736), True),
                ('Automation', 'Thermal Energy', CurricularUnit(418773), False),
                ('Automation', 'Thermal Energy', CurricularUnit(418736), False),
                ('Automation', 'Production Management', CurricularUnit(418773), False),
                ('Automation', 'Production Management', CurricularUnit(418736), False),
                ('Automation', 'Production, Conception and Manufacturing', CurricularUnit(418773), False),
                ('Automation', 'Structural Engineering and Machine Design', CurricularUnit(418736), False),
                ('Automation', None, CurricularUnit(418722), True),
                ('Personal and Interpersonal Skills', 'Automation', CurricularUnit(418787), False),
                ...
            ]
        """

        html = _cache.get_html(self.url, use_cache = use_cache)
        soup = _bs4.BeautifulSoup(html, "lxml")

        for tag in soup.find_all("a"):
            if "cur_geral.cur_planos_estudos_view" in str(tag):
                url = self.base_url + tag["href"]
                break
        else:
            raise ValueError("No study plan link was found on the course page")
        
        html = _cache.get_html(url, use_cache = use_cache)
        soup = _bs4.BeautifulSoup(html, "lxml")

        code_to_uc = {uc.code : uc for uc in self.curricular_units(use_cache = use_cache)}
        result = []

        tables = soup.find_all("table", {"class" : "dados"})

        for table in tables:
            for row in table.find_all("tr")[1:]: # Ignore the header row
                
                keyword, branch, code, acronym, name, _credits, mandatory = (tag.string for tag in row.find_all("td"))

                try:
                    result.append((keyword, branch, code_to_uc[code], "Yes" in str(mandatory)))
                except KeyError: # The curricular unit doesn't have a page
                    continue
        
        return result

    def exams(self, use_cache : bool = True) -> list:
        """Returns a list of the exams of this course.
        
        Args:
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra

        Returns:
            A list of dictionaries (see :func:`exams.exams` for more information about the dictionaries)
        """
        res = []
        
        for faculty_url in self.involved_organic_units:
            url = faculty_url.replace("/en/", "/pt/") +\
                  _utils.SIG_URLS["course exams"] + "?" +\
                  _urllib.parse.urlencode({"p_curso_id" : str(self.pv_curso_id)})
            
            res.extend(_exams.exams(url, use_cache, faculty_url))

        return res


    @classmethod
    def from_url(cls, url : str, use_cache : bool = True, base_url : str = "https://sigarra.up.pt/feup/en/"):
        """Scrapes the course webpage from the given url and returns a :obj:`Course` object.

        Args:
            url (str): The url of the course's sigarra page
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra
            base_url (:obj:`str`, optional): The url of the faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")
        
        Returns:
            A :obj:`Course` object

        Example::
        
            from feupy import Course

            url = "https://sigarra.up.pt/feup/pt/cur_geral.cur_view?pv_curso_id=742&pv_ano_lectivo=2018"
            mieic = Course.from_url(url)

            print(mieic.name)
            # Master in Informatics and Computing Engineering
        """

        try:
            pv_curso_id    = int(_re.findall(r"pv_curso_id=(\d+)"   , url)[0])
            pv_ano_lectivo = int(_re.findall(r"pv_ano_lectivo=(\d+)", url)[0])
        except:
            raise ValueError(f"from_url() 'url' argument \"{url}\" is not a valid course url")

        matches = _re.findall(r"^https?://sigarra\.up\.pt/(\w+)/", url)
        if len(matches) == 1:
            base_url = f"https://sigarra.up.pt/{matches[0]}/en/"

        return Course(pv_curso_id, pv_ano_lectivo, use_cache, base_url = base_url)


    @classmethod
    def from_a_tag(cls, bs4_tag : _bs4.Tag, use_cache : bool = True, base_url : str = "https://sigarra.up.pt/feup/en/"):
        """Scrapes the course webpage from the given :obj:`bs4.tag` object and returns a :obj:`Course` object.
        
        Args:
            bs4_tag (:obj:`bs4.tag`):
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra
            base_url (:obj:`str`, optional): The url of the faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")
        
        Returns:
            A :obj:`Course` object
        """

        if bs4_tag.name != "a":
            raise ValueError(f"from_a_tag() 'bs4_tag' argument must be an anchor tag, not '{bs4_tag.name}'")
        
        return Course.from_url(bs4_tag["href"], use_cache, base_url = base_url)

    
    # Comparisons between curricular units are made with the pv_curso_id and pv_ano_lectivo
    def _unique_value(self):
        return self.pv_curso_id * 10000 + self.pv_ano_lectivo

    def __eq__(self, other):
        if isinstance(other, Course):
            return self._unique_value() == other._unique_value()
        else:
            return NotImplementedError
        
    def __gt__(self, other):
        if isinstance(other, Course):
            return self._unique_value() > other._unique_value()
        else:
            return NotImplementedError
    
    def __ge__(self, other):
        if isinstance(other, Course):
            return self._unique_value() >= other._unique_value()
        else:
            return NotImplementedError
    
    def __lt__(self, other):
        if isinstance(other, Course):
            return self._unique_value() < other._unique_value()
        else:
            return NotImplementedError
    
    def __le__(self, other):
        if isinstance(other, Course):
            return self._unique_value() <= other._unique_value()
        else:
            return NotImplementedError
    
    @property
    def __dict__(self): # This is done for compatibility reasons (vars)
        return {attribute : getattr(self, attribute) for attribute in self.__slots__}

    def __hash__(self):
        return hash(self._unique_value())
    
    def __repr__(self):
        if self.base_url == "https://sigarra.up.pt/feup/en/":
            return f"Course({self.pv_curso_id}, {self.pv_ano_lectivo})"
        else:
            return f"Course({self.pv_curso_id}, {self.pv_ano_lectivo}, base_url = {self.base_url})"
    
    def __str__(self):
        return f"{self.acronym} ({self.pv_ano_lectivo}/{self.pv_ano_lectivo + 1})"



import re as _re
import urllib as _urllib

import bs4 as _bs4

from . import _Credentials
from . import _internal_utils as _utils
from . import _Course
from . import cache as _cache

__all__ = ["Student"]

class Student:
    """This class represents a FEUP student as seen from their sigarra webpage.

    Args:
        username (int): The username of the student, e.g. 201806185
        use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra
        base_url (:obj:`str`, optional): The url of the faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")

    
    Attributes:
        name             (str): The name of the student
        links            (tuple(str)): Urls from the student page (including :any:`Student.personal_webpage`, if present)
        personal_webpage (str): Url of the student's personal page, if present. Otherwise it is set to None
        username         (int): The student's "pv_num_unico"
        url              (str): Url of the student's sigarra page
        base_url         (str): The url of the student's faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")
        courses          (tuple(dict)): The courses this student is enrolled in
    
        Each dictionary from the courses tuple has 3 keys:
            - "course"              (a :obj:`Course` object or a string (if a link to a course wasn't available))
            - "first academic year" (int): if your first year is 2019/2020, then "first academic year" will be 2019
            - "institution"         (string)
    
    Example::
        
        from feupy import Student

        daniel = Student(201806185)

        print(daniel.name)
        # Daniel Filipe Amaro Monteiro

        print(daniel.username)
        # 201806185

        print(daniel.courses)
        # ({'course': Course(742, 2019), 'institution': 'Faculty of Engineering', 'first academic year': 2018},)
    """

    __slots__ = ["name", "links", "personal_webpage", "username", "courses", "url", "base_url"]

    def __init__(self, username : int, use_cache : bool = True, base_url : str = "https://sigarra.up.pt/feup/en/"):

        self.username = username
        self.base_url = base_url
        self.url = self.base_url + _utils.SIG_URLS["student page"] + "?" + _urllib.parse.urlencode({"pv_num_unico" : str(username)})
        
        html = _cache.get_html(url = self.url, use_cache = use_cache) # Getting the html
        soup = _bs4.BeautifulSoup(html, "lxml")

        if "Estudante não encontrado." in html:
            raise ValueError(f"Student with username '{username}' doesn't exist")
        
        if "Problem found" in html: # the only info we can get is the name (I don't think it's even the entire name)
            content_soup = soup.find("div", {"id" : "conteudoinner"})
            self.name = content_soup.contents[5].string

            self.links = () # empty tuple
            self.personal_webpage = None
            self.courses = ()

            return
        
        # Otherwise, it's a normal student page

        self.name = soup.find("div", {"class" : "estudante-info-nome"}).string.strip()
        
        personal_webpage_div = soup.find("div", {"class" : "pagina-pessoal"})
        if personal_webpage_div == None:                              # Does the student have a webpage?
            self.personal_webpage = None
        else:
            self.personal_webpage = personal_webpage_div.a["href"]
        
        self.links = []
        username_link = soup.find("div", {"class" : "estudante-info-numero"})
        for div in username_link.find_next_siblings("div"):
            link = div.a
            if link == None:
                continue
            else:
                self.links.append(link["href"])
        self.links = tuple(self.links)
        
        self.courses = []
        for course_div in soup.find_all("div", {"class" : "estudante-lista-curso-activo"}): # Iterate over the courses "boxes"

            name_div = course_div.find("div", {"class" : "estudante-lista-curso-nome"})
            if name_div.a == None: # There is no link
                course = name_div.string
            else:
                course = _Course.Course.from_a_tag(name_div.a, use_cache, base_url = base_url) # If there is a link, get the Course object
            
            institution = course_div.find("div", {"class" : "estudante-lista-curso-instit"}).string

            first_academic_year = _utils.parse_academic_year(course_div)

            self.courses.append({"course" : course, "institution" : institution, "first academic year" : first_academic_year})
        self.courses = tuple(self.courses)

    @classmethod
    def from_url(cls, url : str, use_cache : bool = True, base_url : str = "https://sigarra.up.pt/feup/en/"):
        """Scrapes the student webpage from the given url and returns a :obj:`Student` object.

        Args:
            url (str): The url of the student's sigarra page
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra
            base_url (:obj:`str`, optional): The url of the faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")
        
        Returns:
            A :obj:`Student` object

        Example::
        
            from feupy import Student

            url = "https://sigarra.up.pt/feup/pt/fest_geral.cursos_list?pv_num_unico=201806185"
            daniel = Student.from_url(url)

            print(daniel.name)
            # Daniel Filipe Amaro Monteiro
        """
        
        matches = _re.findall(r"pv_num_unico=(\d+)$", url)
        
        if len(matches) == 0:
            raise ValueError(f"from_url() 'url' argument \"{url}\" is not a valid student url")
        
        username = int(matches[0])

        matches = _re.findall(r"^https?://sigarra\.up\.pt/(\w+)/", url)
        if len(matches) == 1:
            base_url = f"https://sigarra.up.pt/{matches[0]}/en/"

        return Student(username, use_cache, base_url = base_url)
    
    @classmethod
    def from_a_tag(cls, bs4_tag : _bs4.Tag, use_cache : bool = True, base_url : str = "https://sigarra.up.pt/feup/en/"):
        """Scrapes the student webpage from the given :obj:`bs4.tag` object and returns a :obj:`Student` object.
        
        Args:
            bs4_tag (:obj:`bs4.tag`):
            use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra
            base_url (:obj:`str`, optional): The url of the faculty (in english) (defaults to "https://sigarra.up.pt/feup/en/")
        
        Returns:
            A :obj:`Student` object
        """
        
        if not isinstance(bs4_tag, _bs4.Tag):
            raise TypeError(f"from_a_tag() 'bs4_tag' argument must be a bs4.Tag, not '{type(bs4_tag).__name__}'")
        
        if bs4_tag.name != "a":
            raise ValueError(f"from_a_tag() 'bs4_tag' argument must be an anchor tag, not '{bs4_tag.name}'")
        
        return Student.from_url(bs4_tag["href"], use_cache, base_url = base_url)
    
    def full_info(self, credentials : _Credentials.Credentials) -> dict:
        """Returns a dictionary with the information that one can get when it is logged in.

        The dictionary has 7 keys:
            - "courses" (list(dict))
            - "email"   (str)
            - "links"   (tuple(str))
            - "name"    (str)
            - "personal_webpage" (str)
            - "url"     (str)
            - "username" (int)

        Args:
            credentials (:obj:`Credentials`): A :obj:`Credentials` object
        
        Returns:
            A dictionary

        Example::

            from feupy import Student, Credentials
            from pprint import pprint

            daniel = Student(201806185)

            creds = Credentials()

            pprint(daniel.full_info(creds))

            # You will get something like this:
            {'courses': [ # A list of dictionaries representing the courses' information
                         {'course': Course(742, 2019), # A Course object. If a link to an object isn't available, it's just a string
                          'current year': 2,   # Could be None if a number isn't present (or couldn't be parsed)
                          'first academic year': 2018,
                          'institution': 'Faculty of Engineering', # Best faculty
                          'status': 'A Frequentar'}],
            'email': 'up201806185@fe.up.pt',
            'links': (),                      # personal_webpage is included in links
            'name': 'Daniel Filipe Amaro Monteiro', # people tend to have a name
            'personal_webpage': None,
            'url': 'https://sigarra.up.pt/feup/en/fest_geral.cursos_list?pv_num_unico=201806185',
            'username': 201806185}
        """

        if not isinstance(credentials, _Credentials.Credentials):
            raise TypeError(f"full_info() 'credentials' argument must be a Credentials object, not '{type(credentials).__name__}'")
        
        info = {
            "name"             : self.name,
            "links"            : self.links,
            "personal_webpage" : self.personal_webpage,
            "username"         : self.username,
            "url"              : self.url
        }

        html = credentials.get_html(self.url)
        soup = _bs4.BeautifulSoup(html, "lxml")

        email_div = soup.find("div", {"class" : "email-institucional"})
        email = email_div.a.contents[0] + "@" + email_div.a.contents[-1]
        info["email"] = email

        courses = [_parse_course_box(div) for div in soup.find_all("div", {"class" : "estudante-lista-curso-activo"})]
        info["courses"] = courses

        return info
    
    # Comparisons between students are made with the username
    def __eq__(self, other):
        if isinstance(other, Student):
            return self.username == other.username
        else:
            return NotImplementedError
        
    def __gt__(self, other):
        if isinstance(other, Student):
            return self.username > other.username
        else:
            return NotImplementedError
    
    def __ge__(self, other):
        if isinstance(other, Student):
            return self.username >= other.username
        else:
            return NotImplementedError
    
    def __lt__(self, other):
        if isinstance(other, Student):
            return self.username < other.username
        else:
            return NotImplementedError
    
    def __le__(self, other):
        if isinstance(other, Student):
            return self.username <= other.username
        else:
            return NotImplementedError
    
    @property
    def __dict__(self): # This is done for compatibility reasons (vars)
        return {attribute : getattr(self, attribute) for attribute in self.__slots__}

    def __hash__(self):
        return hash(self.username)
    
    def __repr__(self):
        if self.base_url == "https://sigarra.up.pt/feup/en/":
            return f"Student({self.username})"
        else:
            return f"Student({self.username}, base_url = {self.base_url})"
    
    def __str__(self):
        return f"{self.name} ({self.username})"


def _parse_course_box(bs_course_div):
    """Parses the the information available in the "estudante-lista-curso-activo" div.
    NOTE: Doesn't work with unpriviledged access to the student page
    It returns a dictionary like this: 
    {
        "course" : Course,
        "institution" : str,
        "current year": int, (None if a number couldn't be parsed)
        "status"      : str,
        "first academic year" : int
    }
    """
    
    if not isinstance(bs_course_div, _bs4.BeautifulSoup) and not isinstance(bs_course_div, _bs4.element.Tag):
        raise TypeError(f"parse_course_box() 'bs_course_div' argument must either be a bs4.element.Tag or a bs4.BeautifulSoup, not '{type(bs_course_div).__name__}'")
    
    name_div = bs_course_div.find("div", {"class" : "estudante-lista-curso-nome"})
    if name_div.a == None:  # There is no link
        course = name_div.string
    else:
        course = _Course.Course.from_a_tag(name_div.a)
        
    institution = bs_course_div.find("div", {"class": "estudante-lista-curso-instit"}).string
    
    table = bs_course_div.find("table", {"class" : "formulario"})
    table_data = _utils.scrape_html_table(table)
            
    current_year, status, first_academic_year = [row[1] for row in table_data]
    
    try:
        current_year = int(current_year)
    except Exception:
        current_year = None
        
    # status is fine as it is
    
    first_academic_year = _utils.parse_academic_year(first_academic_year)
    
    result = {
        "course"              : course,
        "institution"         : institution,
        "current year"        : current_year,
        "status"              : status,
        "first academic year" : first_academic_year
    }
    
    return result
    



"""The sigarra scraping library no one asked for

Objects:
    Student
    Teacher
    CurricularUnit
    Course
    Credentials
    User

Modules:
    exams
    timetable
    cache
"""
from . _Student import Student
from . _Teacher import Teacher
from . _CurricularUnit import CurricularUnit
from . _Course import Course

from . _Credentials import Credentials
from . _User import User

from . import exams
from . import timetable


import unittest

from .FeupyTestCase import FeupyTestCase
from feupy import CurricularUnit, Teacher

class TestFpro(FeupyTestCase):
    @classmethod
    def setUpClass(cls):
        cls.fpro = CurricularUnit(419983)

    def test_attributes(self):
        expected_output = {
            'ECTS_credits': 6.0,
            'academic_year': 2018,
            'acronym': 'FPRO',
            'base_url': 'https://sigarra.up.pt/feup/en/',
            'code': 'EIC0005',
            'curricular_years': (1,),
            'has_moodle': True,
            'name': 'Programming Fundamentals',
            'number_of_students': 182,
            'pv_ocorrencia_id': 419983,
            'regents': (Teacher(230756),),
            'semester': '1',
            'teachers': (Teacher(230756),
                        Teacher(230756),
                        Teacher(520205),
                        Teacher(552793),
                        Teacher(209847)),
            'url': 'https://sigarra.up.pt/feup/en/ucurr_geral.ficha_uc_view?pv_ocorrencia_id=419983',
            'webpage_url': 'https://web.fe.up.pt/~jlopes/doku.php/teach/fpro/index',
        }
        
        self.assertObjectAttributes(self.fpro, expected_output)


class TestIope(FeupyTestCase):
    @classmethod
    def setUpClass(cls):
        cls.iope = CurricularUnit(436840)
    
    def test_attributes(self):
        expected_output = {
            'ECTS_credits': 6.0,
            'academic_year': 2019,
            'acronym': 'IOPE',
            'base_url': 'https://sigarra.up.pt/feup/en/',
            'code': 'EEC0017',
            'curricular_years': (3, 4),
            'has_moodle': True,
            'name': 'Operations Research',
            'number_of_students': 148,
            'pv_ocorrencia_id': 436840,
            'regents': (Teacher(209980), Teacher(233753), Teacher(452947)),
            'semester': '1',
            'teachers': (Teacher(209980),
                        Teacher(233753),
                        Teacher(452947),
                        Teacher(468729),
                        Teacher(468729),
                        Teacher(452947),
                        Teacher(209980),
                        Teacher(233753)),
            'url': 'https://sigarra.up.pt/feup/en/ucurr_geral.ficha_uc_view?pv_ocorrencia_id=436840',
            'webpage_url': 'https://sites.google.com/g.uporto.pt/io-or',
            }

        self.assertObjectAttributes(self.iope, expected_output)

class TestPF(FeupyTestCase):
    @classmethod
    def setUpClass(cls):
        cls.pf = CurricularUnit(420521)
    
    def test_attributes(self):
        expected_output = {
            'ECTS_credits': 1.5,
            'academic_year': 2018,
            'acronym': 'PF',
            'base_url': 'https://sigarra.up.pt/feup/en/',
            'code': 'FEUP002',
            'curricular_years': (1,),
            'has_moodle': True,
            'name': 'Project FEUP',
            'number_of_students': 930,
            'pv_ocorrencia_id': 420521,
            'regents': (Teacher(322605), Teacher(245546)),
            'semester': '1',
            'teachers': (
                Teacher(210017), Teacher(239083), Teacher(210893), 
                Teacher(207140), Teacher(209784), Teacher(209810), 
                Teacher(322605), Teacher(210741), Teacher(472718), 
                Teacher(245546), Teacher(211430), Teacher(230756), 
                Teacher(555321), Teacher(210176), Teacher(208327), 
                Teacher(207140), Teacher(249597), Teacher(246626), 
                Teacher(209784), Teacher(209810), Teacher(500614), 
                Teacher(210741), Teacher(208947), Teacher(472718), 
                Teacher(211430), Teacher(249340), Teacher(243027), 
                Teacher(400655), Teacher(210158), Teacher(209821), 
                Teacher(211795), Teacher(210660), Teacher(209876), 
                Teacher(384641), Teacher(235777)
                ),
            'url': 'https://sigarra.up.pt/feup/en/ucurr_geral.ficha_uc_view?pv_ocorrencia_id=420521',
            'webpage_url': 'http://www.fe.up.pt/projfeup~',
            }

        self.assertObjectAttributes(self.pf, expected_output)

class TestPDMPAThesis(FeupyTestCase):
    @classmethod
    def setUpClass(cls):
        cls.thesis = CurricularUnit(436365)
    
    def test_attributes(self):
        expected_output = {
                'ECTS_credits': 198.0,
                'academic_year': 2019,
                'acronym': 'TESE',
                'base_url': 'https://sigarra.up.pt/feup/en/',
                'code': 'PDMPA31',
                'curricular_years': (1, 2, 3, 4),
                'has_moodle': False,
                'name': 'Thesis',
                'number_of_students': 8,
                'pv_ocorrencia_id': 436365,
                'regents': (Teacher(209625),),
                'semester': 'A',
                'teachers': (Teacher(207878),
                            Teacher(207894),
                            Teacher(231674),
                            Teacher(245140),
                            Teacher(424026),
                            Teacher(209625)),
                'url': 'https://sigarra.up.pt/feup/en/ucurr_geral.ficha_uc_view?pv_ocorrencia_id=436365',
                'webpage_url': None,
            }

        self.assertObjectAttributes(self.thesis, expected_output)

if __name__ == '__main__':
    unittest.main()



import datetime as _datetime

import bs4 as _bs4

from . import cache as _cache
from . import _CurricularUnit

__all__ = ["exams"]

def exams(url : str, use_cache : bool = True, base_url : str = "https://sigarra.up.pt/feup/en/") -> list:
    """Returns a list of dictionaries.

    Each dictionary represents an exam and has 6 keys:

    =================== ==========================================================
     key                 value                                                    
    =================== ==========================================================
     "curricular unit"   (:obj:`feupy.CurricularUnit`) The subject being evaluated      
     "finish"            (:obj:`datetime.datetime`) The finish time of the exam   
     "observations"      (str)                                                    
     "rooms"             (tuple(str)) The rooms in which the exam will take place 
     "season"            (str)                                                    
     "start"             (:obj:`datetime.datetime`) The start time of the exam    
    =================== ==========================================================
    
    Note:
        If a value is not available, it will be set to None.
    
    Args:
        url       (str): The url of the page with the exams to be parsed
        use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra

    Returns:
        A list of dictionaries
    
    Example::

        from feupy.exams import exams
        from pprint import pprint

        mieic_exams_url = "https://sigarra.up.pt/feup/pt/exa_geral.mapa_de_exames?p_curso_id=742"

        pprint(exams(mieic_exams_url))
        # You'll get something like this:
        [{'curricular unit': CurricularUnit(420016),
        'finish': datetime.datetime(2019, 9, 4, 12, 0),
        'observations': 'Exame em comum com Mecanica.Exame em comum com o da Época '
                        'de Conclusão de Curso',
        'rooms': ('B222',),
        'season': 'Exames ao abrigo de estatutos especiais - Port.Est.Especiais 2ºS',
        'start': datetime.datetime(2019, 9, 4, 9, 0)},

        {'curricular unit': CurricularUnit(419990),
        'finish': datetime.datetime(2019, 9, 6, 17, 0),
        'observations': 'Sala de exame - Em principio 2 alunos ',
        'rooms': ('B222',),
        'season': 'Exames ao abrigo de estatutos especiais - Port.Est.Especiais 2ºS',
        'start': datetime.datetime(2019, 9, 6, 14, 0)},

        {'curricular unit': CurricularUnit(420021),
        'finish': datetime.datetime(2019, 9, 18, 17, 30),
        'observations': 'Sala de exame - possivelmente 3 alunos ',
        'rooms': None,
        'season': 'Exames ao abrigo de estatutos especiais - Port.Est.Especiais 2ºS',
        'start': datetime.datetime(2019, 9, 18, 14, 30)},

        {'curricular unit': CurricularUnit(438941),
        'finish': datetime.datetime(2019, 9, 25, 13, 0),
        'observations': None,
        'rooms': ('B104', 'B208', 'B213'),
        'season': 'Exames ao abrigo de estatutos especiais - Mini-testes (1ºS)',
        'start': datetime.datetime(2019, 9, 25, 9, 0)},

        {'curricular unit': CurricularUnit(438941),
        'finish': datetime.datetime(2019, 9, 25, 17, 30),
        'observations': None,
        'rooms': ('B104', 'B213', 'B208', 'B207'),
        'season': 'Exames ao abrigo de estatutos especiais - Mini-testes (1ºS)',
        'start': datetime.datetime(2019, 9, 25, 13, 30)},

        {'curricular unit': CurricularUnit(438941),
        'finish': datetime.datetime(2019, 9, 26, 13, 0),
        'observations': None,
        'rooms': ('B104', 'B208', 'B213'),
        'season': 'Exames ao abrigo de estatutos especiais - Mini-testes (1ºS)',
        'start': datetime.datetime(2019, 9, 26, 9, 0)},

        {'curricular unit': CurricularUnit(438941),
        'finish': datetime.datetime(2019, 9, 26, 17, 30),
        'observations': None,
        'rooms': ('B104', 'B213', 'B208', 'B207'),
        'season': 'Exames ao abrigo de estatutos especiais - Mini-testes (1ºS)',
        'start': datetime.datetime(2019, 9, 26, 13, 30)}]
    """
    html = _cache.get_html(url, use_cache = use_cache)
    soup = _bs4.BeautifulSoup(html, 'lxml')
    content = soup.find("div", {"id" : "conteudoinner"})

    tags = [tag for tag in content.find_all("a")]
    tags.pop(0)

    urls = [base_url.replace("/en/", "/pt/") + tag["href"] for tag in tags if "exa_geral.exame_view" in tag["href"]]
    _cache.get_html_async(urls, use_cache = use_cache) # Refresh the cache
    return [_parse_exam_page(url) for url in urls]

def _parse_exam_page(url : str):
    """Parses an exam page from the url and returns a dictionary"""

    html = _cache.get_html(url)
    soup = _bs4.BeautifulSoup(html, 'lxml')
    content = soup.find("div", {"id" : "conteudoinner"})
    table = content.find("table")

    rows = table.find_all("tr")

    code, _, season, date, start, length = rows[:6]

    curricular_unit = _CurricularUnit.CurricularUnit.from_a_tag(code.a)

    season = season.find_all("td")[1].string

    year, month, day = map(int, date.find_all("td")[1].string.split("-"))

    start_hour, start_minute = map(int, start.find_all("td")[1].string.split(":"))

    start = _datetime.datetime(year, month, day, start_hour, start_minute)

    try:
        length_hour, length_minute = map(int, length.find_all("td")[1].string.split(":"))

        delta = _datetime.timedelta(hours = length_hour, minutes = length_minute)

        finish = start + delta

    except AttributeError:
        finish = None

    rooms = observations = None
    for row in rows[6:]:
        if "Salas:" in str(row):
            rooms = tuple(tag.string for tag in row.find_all("a"))
        elif "Observações:" in str(row):
            observations = row.find_all("td")[1].string
    
    return {
        "curricular unit" : curricular_unit,
        "season"          : season,
        "start"           : start,
        "finish"          : finish,
        "rooms"           : rooms,
        "observations"    : observations
    }



import re as _re

import bs4 as _bs4

from . import _Credentials
from . import _internal_utils as _utils
from . import _Course
from . import _CurricularUnit
from . import timetable as _timetable

__all__ = ["User"]

class User:
    """This class represents the information that can be extracted from your personal webpage.
    
    Args:
        pv_fest_id (int): Your pv_fest_id. See :func:`User.get_pv_fest_ids` for a way
            to get this value. If you are enrolled in only one course, then
            :func:`User.from_credentials` may be more convenient for you
        
        credentials (:obj:`Credentials`): A :obj:`Credentials` object

    Attributes:
        pv_fest_id  (int): A sort of course specific student identification number
        course      (:obj:`Course`): The course related to this pv_fest_id
        credentials (:obj:`Credentials`): A :obj:`Credentials` object
            (This is done to avoid having to pass the same :obj:`Credentials` object to every function)

    Example::

        from feupy import Credentials, User

        creds = Credentials()

        me = User(123456, creds)
    """

    __slots__ = ["pv_fest_id", "course", "credentials"]

    def __init__(self, pv_fest_id : int, credentials: _Credentials.Credentials):
        
        self.pv_fest_id = pv_fest_id
        self.credentials = credentials # Note, this is only a reference

        html = credentials.get_html(self.credentials.base_url + _utils.SIG_URLS["courses units"], {"pv_fest_id" : str(pv_fest_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")
        
        if "Não tem permissões para aceder a este conteúdo." in html:
            raise ValueError("Your Credentials object does not have permission to access the page related to this pv_fest_id")

        for tag in soup.find_all("h2"):
            if "cur_geral.cur_view" in str(tag):
                self.course = _Course.Course.from_a_tag(tag.a)
                break
        else:
            raise Exception("No course was found")

    def courses_units(self) -> list:
        """Returns your grades as a list of tuples.
        
        Each tuple has 2 elements:
            0. A :obj:`CurricularUnit` object which represents a curricular unit that
                you either have had in a previous year or you are currently enrolled in
            1. Either an int which represents the grade you got at that curricular unit or None
                if that grade is not available

        Returns:
            A list of tuples

        Example::

            from feupy import Credentials, User
            from pprint import pprint


            creds = Credentials()
            me = User.from_credentials(creds)

            pprint(me.courses_units())

            # You will get something like this:
            [(CurricularUnit(419981), 10),
             (CurricularUnit(419982), 11),
             (CurricularUnit(419983), 12),
             (CurricularUnit(419984), 13),
             (CurricularUnit(419985), 14),
             (CurricularUnit(420521), 15),
             (CurricularUnit(419986), None),
             (CurricularUnit(419987), None),
             (CurricularUnit(419988), None),
             (CurricularUnit(419989), None),
             (CurricularUnit(419990), None)]
        """
        html = self.credentials.get_html(self.credentials.base_url + _utils.SIG_URLS["courses units"], {"pv_fest_id" : str(self.pv_fest_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")

        result = []
        for row in soup.find_all("tr", {"class" : "d"}):
            curricular_unit = _CurricularUnit.CurricularUnit.from_a_tag(row.a)
            
            try:
                grade = int(row.find("td", {"class" : "n"}).string)
            except:
                grade = None
            
            result.append((curricular_unit, grade))
        
        return result

    def timetable(self) -> list:
        """Returns the current user timetable as a list of dictionaries if possible, otherwise returns None.
        (see :func:`timetable.parse_current_timetable` for more info)
        
        Returns:
            A list of dicts
        """
        html = self.credentials.get_html(self.credentials.base_url.replace("/en/", "/pt/") + _utils.SIG_URLS["personal timetable"], {"pv_fest_id" : str(self.pv_fest_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")

        return _timetable.parse_current_timetable(self.credentials, soup.a["href"])
    
    def all_timetables(self) -> dict:
        """Parses all the timetables related to this user
        (see :obj:`timetable.parse_timetables` for further info).
        
        Returns:
            A dictionary which maps a tuple with two :obj:`datetime.date` objects,
            start and finish (the time span in which this timetable is valid), to a
            list of dictionaries (see :obj:`timetable.parse_timetable` for an example
            of such a list).
        """
        html = self.credentials.get_html(self.credentials.base_url.replace("/en/", "/pt/") + _utils.SIG_URLS["personal timetable"], {"pv_fest_id" : str(self.pv_fest_id)})
        soup = _bs4.BeautifulSoup(html, "lxml")

        return _timetable.parse_timetables(self.credentials, soup.a["href"])

    def classes(self):
        """Returns the classes you are in as a list of tuples.
        
        Each tuple has 2 elements:
            0. A :obj:`CurricularUnit` object
            1. The class name as a string.
        
        Returns:
            A list of tuples
        """
        raise Warning("No idea if this works or not")
        html = self.credentials.get_html(self.credentials.base_url + _utils.SIG_URLS["classes data"] , params = {"pv_estudante_id" : str(self.pv_fest_id)})
        soup = _bs4.BeautifulSoup(html, 'lxml')
    
        tables = soup.find_all("table", {"class" : "tabela"})[1:] # Forget first table

        result = []
        for table in tables:
            for row in table.find_all("tr")[2:]: #Forget the header rows
                curricular_unit = _CurricularUnit.CurricularUnit.from_a_tag(row.find("a"))
                class_name = row.find_all("td")[3].string

                result.append((curricular_unit, class_name))
        
        return result

    @classmethod
    def from_credentials(cls, credentials: _Credentials.Credentials):
        """Returns a User object made from the first result from get_pv_fest_ids().

        Usually, students are enrolled in only one course, which means that
        get_pv_fest_ids() tends to be a tuple with a single int.

        Args:
            credentials (:obj:`Credentials`): A :obj:`Credentials` object
        
        Returns:
            A :obj:`User` object
        
        Example::

            from feupy import Credentials, User

            creds = Credentials()
            me = User.from_credentials(creds)
        """
        return User(User.get_pv_fest_ids(credentials)[0], credentials)


    @staticmethod
    def get_pv_fest_ids(credentials: _Credentials.Credentials) -> tuple:
        """Returns a tuple of ints, each representing a pv_fest_id.
        
        Args:
            credentials (:obj:`Credentials`): A :obj:`Credentials` object
        
        Returns:
            A tuple of ints
        """
        payload = {"pv_num_unico" : str(credentials.username)}
        html = credentials.get_html(credentials.base_url + _utils.SIG_URLS["student page"], payload)
    
        matches = _re.findall(r"pv_fest_id=(\d+)", html)

        if len(matches) == 0:
            raise ValueError("No pv_fest_id's were found")
    
        return tuple(int(match) for match in matches)

    @property
    def __dict__(self): # This is done for compatibility reasons (vars)
        return {attribute : getattr(self, attribute) for attribute in self.__slots__}



import datetime as _datetime
import functools as _functools
import re as _re

import bs4 as _bs4

from . import _Credentials
from . import _CurricularUnit
from . import _Teacher

__all__ = ["parse_current_timetable", "parse_timetables", "parse_timetable"]

_weekdays = ("Monday", "Tuesday", "Wednesday",
             "Thursday", "Friday", "Saturday", "Sunday")

_weekdays_pt = ("Segunda", "Terça", "Quarta",
                "Quinta", "Sexta", "Sábado", "Domingo")

_weekdays_pt_to_en = {pt: en for pt, en in zip(_weekdays_pt, _weekdays)}


def parse_current_timetable(credentials: _Credentials.Credentials, url: str):
    """Attempts to return the related timetable that is valid today as a list of dictionaries.
    If no timetable is valid today, it returns None.
    
    Args:
        credentials (:obj:`feupy.Credentials`): A :obj:`feupy.Credentials` object
        url (str): The url of the timetable page

    Returns:
        A list of dicts (e.g. see :func:`parse_timetable`) or None

    """
    
    for (start, finish), url in _parse_side_bar(credentials, url).items():
        if start <= _datetime.date.today() <= finish:
            return parse_timetable(credentials, url)
    else:
        return None


def parse_timetables(credentials: _Credentials.Credentials, url: str) -> dict:
    """Returns the timetables related to this timetable (including) as a dictionary.

    Args:
        credentials (:obj:`feupy.Credentials`): A :obj:`feupy.Credentials` object
        url (str): The url of the timetable page

    Returns:
        A dictionary which maps a tuple with two :obj:`datetime.date` objects,
        start and finish (the time span in which this timetable was/is valid), to a
        list of dictionaries (e.g. see :func:`parse_timetable`).

    Example::

        from feupy import timetable, Credentials
        from pprint import pprint

        timetable_url = "https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=207783&pv_periodos=1&pv_ano_lectivo=2017"
        creds = Credentials()

        pprint(timetable.parse_timetables(creds, timetable_url))
        # You'll get something like this:
        {(datetime.date(2017, 9, 24), datetime.date(2017, 10, 28)): [{'class type': 'TP',
                                                                    'classes': ('1MIEIC01',),
                                                                    'curricular unit': CurricularUnit(399800),
                                                                    'finish': datetime.time(11, 0),
                                                                    'room': ('B232A',),
                                                                    'start': datetime.time(9, 0),
                                                                    'teachers': (Teacher(212345),),
                                                                    'weekday': 'Monday'},
                                                                    {'class type': 'T',
                                                                    'classes': ('1MIEIC01',
                                                                                '1MIEIC02',
                                                                                '1MIEIC03',
                                                                                '1MIEIC04',
                                                                                '1MIEIC05',
                                                                                '1MIEIC06',
                                                                                '1MIEIC07',
                                                                                '1MIEIC08'),
                                                                    'curricular unit': CurricularUnit(399800),
                                                                    'finish': datetime.time(13, 0),
                                                                    'room': ('B003',),
                                                                    'start': datetime.time(11, 0),
                                                                    'teachers': (Teacher(212345),),
                                                                    'weekday': 'Monday'},
                                                                    {'class type': 'T',
                                                                    'classes': ('1MIEIC01',
                                                                                '1MIEIC02',
                                                                                '1MIEIC03',
                                                                                '1MIEIC04',
                                                                                '1MIEIC05',
                                                                                '1MIEIC06',
                                                                                '1MIEIC07',
                                                                                '1MIEIC08'),
                                                                    'curricular unit': CurricularUnit(399800),
                                                                    'finish': datetime.time(10, 0),
                                                                    'room': ('B001',),
                                                                    'start': datetime.time(8, 30),
                                                                    'teachers': (Teacher(212345),
                                                                                Teacher(212345)),
                                                                    'weekday': 'Tuesday'},
                                                                    {'class type': 'T',
                                                                    'classes': ('1MIEIC01',
                                                                                '1MIEIC02',
                                                                                '1MIEIC03',
                                                                                '1MIEIC04',
                                                                                '1MIEIC05',
                                                                                '1MIEIC06',
                                                                                '1MIEIC07',
                                                                                '1MIEIC08'),
                                                                    'curricular unit': CurricularUnit(399800),
                                                                    'finish': datetime.time(11, 30),
                                                                    'room': ('B001',),
                                                                    'start': datetime.time(10, 0),
                                                                    'teachers': (Teacher(212345),),
                                                                    'weekday': 'Tuesday'},
                                                                    {'class type': 'TP',
                                                                    'classes': ('1MIEIC02',),
                                                                    'curricular unit': CurricularUnit(399800),
                                                                    'finish': datetime.time(13, 30),
                                                                    'room': ('B110',),
                                                                    'start': datetime.time(11, 30),
                                                                    'teachers': (Teacher(212345),),
                                                                    'weekday': 'Tuesday'},
                                                                    {'class type': 'T',
                                                                    'classes': ('1MIEIC01',
                                                                                '1MIEIC02',
                                                                                '1MIEIC03',
                                                                                '1MIEIC04',
                                                                                '1MIEIC05',
                                                                                '1MIEIC06',
                                                                                '1MIEIC07',
                                                                                '1MIEIC08'),
                                                                    'curricular unit': CurricularUnit(399800),
                                                                    'finish': datetime.time(10, 0),
                                                                    'room': ('B001',),
                                                                    'start': datetime.time(8, 30),
                                                                    'teachers': (Teacher(212345),),
                                                                    'weekday': 'Wednesday'},
                                                                    {'class type': 'T',
                                                                    'classes': ('1MIEIC01',
                                                                                '1MIEIC02',
                                                                                '1MIEIC03',
                                                                                '1MIEIC04',
                                                                                '1MIEIC05',
                                                                                '1MIEIC06',
                                                                                '1MIEIC07',
                                                                                '1MIEIC08'),
                                                                    'curricular unit': CurricularUnit(399800),
                                                                    'finish': datetime.time(11, 30),
                                                                    'room': ('B001',),
                                                                    'start': datetime.time(10, 0),
                                                                    'teachers': (Teacher(212345),
                                                                                Teacher(212345)),
                                                                    'weekday': 'Wednesday'},
                                                                    {'class type': 'TP',
                                                                    'classes': ('1MIEIC03',),
                                                                    'curricular unit': CurricularUnit(399800),
                                                                    'finish': datetime.time(13, 30),
                                                                    'room': ('B205',),
                                                                    'start': datetime.time(11, 30),
                                                                    'teachers': (Teacher(212345),),
                                                                    'weekday': 'Wednesday'},
                                                                    ... ],
        (datetime.date(2017, 10, 29), datetime.date(2017, 11, 4)): [{'class type': 'TP',
                                                                    'classes': ('1MIEIC04',),
                                                                    'curricular unit': CurricularUnit(399800),
                                                                    'finish': datetime.time(11, 0),
                                                                    'room': ('B232A',),
                                                                    'start': datetime.time(9, 0),
                                                                    'teachers': (Teacher(212345),),
                                                                    'weekday': 'Monday'},
                                                                    {'class type': 'T',
                                                                    'classes': ('1MIEIC01',
                                                                                '1MIEIC02',
                                                                                '1MIEIC03',
                                                                                '1MIEIC04',
                                                                                '1MIEIC05',
                                                                                '1MIEIC06',
                                                                                '1MIEIC07',
                                                                                '1MIEIC08'),
                                                                    'curricular unit': CurricularUnit(399800),
                                                                    'finish': datetime.time(13, 0),
                                                                    'room': ('B003',),
                                                                    'start': datetime.time(11, 0),
                                                                    'teachers': (Teacher(212345),),
                                                                    'weekday': 'Monday'},
                                                                    {'class type': 'T',
                                                                    'classes': ('1MIEIC01',
                                                                                '1MIEIC02',
                                                                                '1MIEIC03',
                                                                                '1MIEIC04',
                                                                                '1MIEIC05',
                                                                                '1MIEIC06',
                                                                                '1MIEIC07',
                                                                                '1MIEIC08'),
                                                                    'curricular unit': CurricularUnit(399800),
                                                                    'finish': datetime.time(10, 0),
                                                                    'room': ('B001',),
                                                                    'start': datetime.time(8, 30),
                                                                    'teachers': (Teacher(212345),
                                                                                Teacher(212345)),
                                                                    'weekday': 'Tuesday'},
                                                                    {'class type': 'T',
                                                                    'classes': ('1MIEIC01',
                                                                                '1MIEIC02',
                                                                                '1MIEIC03',
                                                                                '1MIEIC04',
                                                                                '1MIEIC05',
                                                                                '1MIEIC06',
                                                                                '1MIEIC07',
                                                                                '1MIEIC08'),
                                                                    'curricular unit': CurricularUnit(399800),
                                                                    'finish': datetime.time(11, 30),
                                                                    'room': ('B001',),
                                                                    'start': datetime.time(10, 0),
                                                                    'teachers': (Teacher(212345),),
                                                                    'weekday': 'Tuesday'},
                                                                    {'class type': 'TP',
                                                                    'classes': ('1MIEIC06',),
                                                                    'curricular unit': CurricularUnit(399800),
                                                                    'finish': datetime.time(13, 30),
                                                                    'room': ('B110',),
                                                                    'start': datetime.time(11, 30),
                                                                    'teachers': (Teacher(212345),),
                                                                    'weekday': 'Tuesday'},
                                                                    ...]
        }
    """
    result = _parse_side_bar(credentials, url)

    for key in result:
        result[key] = parse_timetable(credentials, result[key])

    return result

def _parse_side_bar(credentials: _Credentials.Credentials, url: str) -> dict:
    """Returns a dictionary which maps a tuple with two datetime.date objects,
    start and finish, to a url of the corresponding timetable (string)"""
    html = credentials.get_html(url)
    soup = _bs4.BeautifulSoup(html, 'lxml')

    timetables_links_table = soup.find("table", {"class": "horario-semanas ecra"})

    if timetables_links_table == None:
        matches = _re.findall(r"Semanas de (\d\d)-(\d\d)-(\d\d\d\d) a (\d\d)-(\d\d)-(\d\d\d\d)", html)[0]
        start_day, start_month, start_year, finish_day, finish_month, finish_year  = map(int, matches)

        start = _datetime.date(start_year, start_month, start_day)
        finish = _datetime.date(finish_year, finish_month, finish_day)

        return {(start, finish) : url}
    
    academic_year = int(_re.findall(r"pv_ano_lectivo=(\d\d\d\d)", url)[0])

    # Only the rows that have links have timetables that we have to parse
    timetable_tags = timetables_links_table.find_all("a")

    result = {}
    for tag in timetable_tags:
        result[_parse_dates(tag.string, academic_year)] = credentials.base_url.replace("/en/", "/pt/") + tag["href"]
    return result

def parse_timetable(credentials: _Credentials.Credentials, url: str) -> list:
    """Parses the events (including overlaps) of the timetable
    with the given url as a list of dictionaries.
    
    Each dictionary represents an timetable event (a class) and has 8 keys:

    =================== ==========================================================
     key                 value                                                    
    =================== ==========================================================
     "class type"        (str) E.g. "T", "TP", etc.
     "classes"           (tuple(str)) The classes that are being taught
     "curricular unit"   (:obj:`feupy.CurricularUnit`) The subject being teached
     "finish"            (:obj:`datetime.time`) The finish time of the event
     "room"              (tuple(str)) The rooms in which the event will take place
     "start"             (:obj:`datetime.time`) The start time of the event
     "teachers"          (tuple(:obj:`feupy.Teacher`))
     "weekday"           (str)
    =================== ==========================================================
    
    Args:
        credentials (:obj:`feupy.Credentials`): A :obj:`feupy.Credentials` object
        url (str): The url of the timetable page

    Returns:
        A list of dicts
    
    Example::
    
        from feupy import timetable, Credentials
        from pprint import pprint

        timetable_url = "https://sigarra.up.pt/feup/pt/hor_geral.turmas_view?pv_turma_id=207783&pv_periodos=1&pv_ano_lectivo=2017"
        creds = Credentials()

        pprint(timetable.parse_timetable(creds, timetable_url))
        # You'll get something like this:
        [{'class type': 'TP',
        'classes': ('1MIEIC01',),
        'curricular unit': CurricularUnit(399999),
        'finish': datetime.time(11, 0),
        'room': ('B232A',),
        'start': datetime.time(9, 0),
        'teachers': (Teacher(212345),),
        'weekday': 'Monday'},
        {'class type': 'T',
        'classes': ('1MIEIC01',
                    '1MIEIC02',
                    '1MIEIC03',
                    '1MIEIC04',
                    '1MIEIC05',
                    '1MIEIC06',
                    '1MIEIC07',
                    '1MIEIC08'),
        'curricular unit': CurricularUnit(399999),
        'finish': datetime.time(13, 0),
        'room': ('B003',),
        'start': datetime.time(11, 0),
        'teachers': (Teacher(212345),),
        'weekday': 'Monday'},
        {'class type': 'T',
        'classes': ('1MIEIC01',
                    '1MIEIC02',
                    '1MIEIC03',
                    '1MIEIC04',
                    '1MIEIC05',
                    '1MIEIC06',
                    '1MIEIC07',
                    '1MIEIC08'),
        'curricular unit': CurricularUnit(399999),
        'finish': datetime.time(10, 0),
        'room': ('B001',),
        'start': datetime.time(8, 30),
        'teachers': (Teacher(212345), Teacher(212345)),
        'weekday': 'Tuesday'},
        {'class type': 'T',
        'classes': ('1MIEIC01',
                    '1MIEIC02',
                    '1MIEIC03',
                    '1MIEIC04',
                    '1MIEIC05',
                    '1MIEIC06',
                    '1MIEIC07',
                    '1MIEIC08'),
        'curricular unit': CurricularUnit(399999),
        'finish': datetime.time(11, 30),
        'room': ('B001',),
        'start': datetime.time(10, 0),
        'teachers': (Teacher(212345),),
        'weekday': 'Tuesday'},
        {'class type': 'TP',
        'classes': ('1MIEIC02',),
        'curricular unit': CurricularUnit(399999),
        'finish': datetime.time(13, 30),
        'room': ('B110',),
        'start': datetime.time(11, 30),
        'teachers': (Teacher(212345),),
        'weekday': 'Tuesday'},
        ... ]
    """
    try:
        weeks_url_query = re.findall(r"&p_semana_inicio=\d+&p_semana_fim=\d+$", url)[0]
    except:
        weeks_url_query = ""

    html = credentials.get_html(url)
    soup = _bs4.BeautifulSoup(html, 'lxml')

    timetable_soup = soup.find("table", {"class": "horario"})

    if timetable_soup == None:
        raise ValueError("No timetable was found in the soup")

    result = []

    rows = timetable_soup.find_all("tr", recursive=False)[1:]  # Ignore the header

    for i, row in enumerate(rows):
        start_time = _datetime.timedelta(hours=8, minutes=i * 30)

        tds = row.find_all("td", recursive=False)[1:]  # Ignore the hour

        for n, td in enumerate(tds):
            if td["class"] in (["horas"], ["almoco"]):  # It's whitespace
                continue

            # Otherwise we have an event we need to parse

            # This for loop compensates n in order to reflect the actual weekday of the event
            for event in sorted(result, key=lambda a: _weekdays.index(a["weekday"])): # sort the events by weekday
                if event["start"] < _timedelta_to_time(start_time) < event["finish"]:
                    if _weekdays.index(event["weekday"]) <= n:
                        n += 1
            weekday = _weekdays[n]

            class_type = _re.findall(r"\((\w+)\)", str(td))[0]

            finish_time = start_time + _datetime.timedelta(minutes=int(td["rowspan"]) * 30)

            uc_tag, classes_tag, room_tag, teachers_tag = td.find_all("a")

            result.append(
                {
                    "weekday"        : weekday,
                    "class type"     : class_type,
                    "start"          : _timedelta_to_time(start_time),
                    "finish"         : _timedelta_to_time(finish_time),
                    "curricular unit": _CurricularUnit.CurricularUnit.from_a_tag(uc_tag),
                    "classes"        : _parse_classes(credentials, classes_tag),
                    "room"           : _parse_rooms(credentials, room_tag),
                    "teachers"       : _parse_teachers(credentials, teachers_tag)
                }
            )

    last_table = timetable_soup.find_next("table", {"class": "dados"})

    if "Aulas Sobrepostas" in str(last_table):
        rows = last_table("tr", {"class": "d"})

        for row in rows:
            weekday_pt = row.find("td", {"headers": "t2"}).string
            weekday = _weekdays_pt_to_en[weekday_pt]

            hour, minute = map(int, row.find(
                "td", {"headers": "t3"}).string.split(":"))
            start = _datetime.time(hour=hour, minute=minute)

            url = (credentials.base_url.replace("/en/", "/pt/") + row.find("td",
                                          {"headers": "t6"}).a["href"]).lower()

            if "hor_geral.composto_desc" in url:
                temp_html = credentials.get_html(url)
                temp_soup = _bs4.BeautifulSoup(temp_html, 'lxml')
                temp_content = temp_soup.find("div", {"id": "conteudoinner"})
                url = (credentials.base_url.replace("/en/", "/pt/") + temp_content.find_all("a")
                       [1]["href"]).lower()
            else:
                url+=weeks_url_query # For some odd reason, single classes' urls don't include the start and finish weeks, a bug on sigarra's side perhaps?

            for event in parse_timetable(credentials, url):
                if event["weekday"] == weekday and event["start"] == start:
                    result.append(event)
                    break
            else:
                raise Exception  # Something went wrong

    def sort_key(event): # Sort by day and then by hour
        return (_weekdays.index(event["weekday"]), event["start"])

    return sorted(result, key = sort_key)


def _timedelta_to_time(t: _datetime.timedelta) -> _datetime.time:
    return (_datetime.datetime.min + t).time()


def _parse_teachers(credentials: _Credentials.Credentials, a: _bs4.Tag) -> tuple:
    url = (credentials.base_url.replace("/en/", "/pt/") + a["href"]).lower()

    if "hor_geral.composto_doc" in url:
        html = credentials.get_html(url)
        soup = _bs4.BeautifulSoup(html, 'lxml')
        content = soup.find("div", {"id": "conteudoinner"})
        teachers_tags = content.find_all("a")[1:]
        return tuple(_Teacher.Teacher.from_a_tag(tag) for tag in teachers_tags)

    elif "func_geral.formview" in url:
        return (_Teacher.Teacher.from_a_tag(a),)

    else:
        raise Exception(f"unrecognized url: {url}")


def _parse_classes(credentials: _Credentials.Credentials, a: _bs4.Tag) -> tuple:
    url = (credentials.base_url.replace("/en/", "/pt/") + a["href"]).lower()

    if "hor_geral.composto_desc" in url:
        html = credentials.get_html(url)
        soup = _bs4.BeautifulSoup(html, 'lxml')
        content = soup.find("div", {"id": "conteudoinner"})
        classes_tags = content.find_all("a")[1:]
        return tuple(sorted(tag.string for tag in classes_tags))

    elif "hor_geral.turmas_view" in url:
        return (a.string,)

    else:
        raise Exception(f"unrecognized url: {url}")


def _parse_rooms(credentials: _Credentials.Credentials, a: _bs4.Tag) -> tuple:
    url = (credentials.base_url.replace("/en/", "/pt/") + a["href"]).lower()

    if "hor_geral.composto_salas" in url:
        html = credentials.get_html(url)
        soup = _bs4.BeautifulSoup(html, 'lxml')
        content = soup.find("div", {"id": "conteudoinner"})
        rooms_tags = content.find_all("a")[1:]
        return tuple(tag.string for tag in rooms_tags)

    elif "instal_geral.espaco_view" in url:
        return (a.string,)

    else:
        raise Exception(f"unrecognized url: {url}")


def _parse_dates(dates_string, academic_year : int):
    result = []
    for date_str in dates_string.split("a"):
        day, month = map(int, date_str.split("-"))

        if month >= 9:  # 9 -> September
            year = academic_year
        else:
            year = academic_year + 1
        result.append(_datetime.date(year, month, day))

    return tuple(result)



"""Caching utilities

Attributes:
    cache (:obj:`shelve.DbfilenameShelf` or None (Initially, see :func:`load_cache`)): A persistent dictionary-like object whose values are structured in the following way:
        
        | {
        |     url0 : (timeout0, html0),
        |     url1 : (timeout1, html1),
        |     ...
        | }

        In which url is a string, timeout is an int or a float (which represents the "due by date" as seconds since epoch), and html is a string

"""
import atexit as _atexit
import datetime as _datetime
import os as _os
import random as _random
import re as _re
import shelve as _shelve
import time as _time
import urllib as _urllib

import bs4 as _bs4
import requests as _requests
import requests_futures.sessions as _sessions

from . import _internal_utils as _utils

__all__ = ["load_cache", "get_html", "reset", "remove_invalid_entries", "get_html_async"]

cache = None # This will eventually be the cache object
# The cache is a dictionary-like object whose values are structured in the following way:
# {
#     url0 : (timeout0, html0),
#     url1 : (timeout1, html1),
#     ...
# }
# In which url is a string, timeout is an int or a float (which represents the "due by date" as seconds since epoch), and html is a string
#
# When get_html is called, if the url is not in the cache keys (or the key is present but the cache entry is outdated),
# get_html will first check if the url matches any key in custom_treatments (it matches if key(url) returns True).
# If so the cache will call custom_treatments[key](html), instead of default_treatment(html)
#
# This behaviour was implemented so that commonly accessed pages (curricular units, students, teachers)
# can have their htmls even more shrinked than usual (by returning a timeout and the div with the useful information)

_custom_treatments = {}


############################### Custom treatments ################################

def _is_curricular_unit(url):
    return "ucurr_geral.ficha_uc_view" in url 

def _is_teacher(url):
    return "func_geral.formview" in url

def _is_student(url):
    return "fest_geral.cursos_list" in url

def _trim_curricular_unit(html):
    """Returns a heavily trimmed version of the curricular unit html"""
    html = _utils.trim_html(html)

    soup = _bs4.BeautifulSoup(html, "lxml")

    useful_stuff = soup.find("div", {"id" : "envolvente"})

    useful_stuff.find("div", {"id" : "colunaprincipal"}).decompose() # Remove the left side-bar

    return str(useful_stuff)

def _trim_teacher(html):
    """Returns a heavily trimmed version of the teacher html"""
    html = _utils.trim_html(html)

    soup = _bs4.BeautifulSoup(html, "lxml")

    useful_stuff = soup.find("div", {"id" : "conteudo"})

    return str(useful_stuff)

def _trim_student(html):
    """Returns a heavily trimmed version of the student html"""
    html = _utils.trim_html(html)

    soup = _bs4.BeautifulSoup(html, "lxml")

    useful_stuff = soup.find("div", {"id" : "conteudo-extra"})

    return str(useful_stuff)

def _curricular_unit_treatment(html):
    timeout = _time.time() + _random_radioactive_lifetime(_datetime.timedelta(days = 6*30)) # It is extremely unlikely that a uc page from a previous year will ever change
    
    try:
        #If this fails, it's probably a uc from another faculty
        uc_academic_year = _utils.parse_academic_year(html)
    except IndexError:
        return (timeout, html)

    if uc_academic_year < _utils.get_current_academic_year():
        return (timeout, _trim_curricular_unit(html))
    else:
        return (timeout, _trim_curricular_unit(html))

def _teacher_treatment(html):
    timeout = _time.time() + _random_radioactive_lifetime(_datetime.timedelta(days = 2*30))
    return (timeout, _trim_teacher(html))

def _student_treatment(html):
    year  = _utils.get_current_academic_year() + 1
    timeout = _datetime.datetime(year, 9, 20).timestamp() # Only update at the beginning of the next academic year
    return (timeout, _trim_student(html))

############################ End of custom treatments ############################

_custom_treatments = {
    _is_curricular_unit : _curricular_unit_treatment,
    _is_teacher         : _teacher_treatment,
    _is_student         : _student_treatment
}

def load_cache(flag = "c", path = None):
    """Loads the cache from disk and stores it in the variable :data:`cache`. 
    If :data:`cache` is different than None, the function will do nothing.

    Args:
        flag (:obj:`str`, optional): The flag parameter, see https://docs.python.org/3/library/dbm.html#dbm.open
        path (:obj:`str` or :obj:`None`, optional): The path of the directory where the cache is stored.
            It defaults to this file's folder path
    
    Note:
        Unless you intend to call :func:`load_cache` with non-default arguments,
        you don't have to call this function. The other functions in this module
        check whether or not the cache has been loaded and will load the cache for
        you.

    Example::
            
        from feupy import cache
        cache.load_cache()

    """
    global cache

    if cache != None: # There is something already in cache
        return
    
    if path == None:
        path = _os.path.dirname(__file__)
    
    cache = _shelve.open(_os.path.join(path, "cache"), flag = flag, writeback = True)
    
    _atexit.register(cache.close) # This function will be called at program termination, in order to make sure that the cache is saved to disk
    # the cache object itself would most likely close itself on exit, but better be safe than sorry

def _random_radioactive_lifetime(half_life = _datetime.timedelta(days=2), cutoff = True):
    """Returns a randomly generated lifetime of a radioactive particle with a half-life of half_life seconds
    (a time_delta is also acceptable).
    If cutoff is True (which is by default), if the lifetime generated is bigger than half_life * 4, the function
    will instead return half_life * 4, which is useful if you wish to avoid ridiculously large lifetimes
    This function is used to disperse the timeouts of the cache values.
    """
    if isinstance(half_life, _datetime.timedelta):
        half_life = half_life.total_seconds()

    lambd = 0.693147 / half_life # ln(2)/half-life
    lifetime = _random.expovariate(lambd)

    return min(lifetime, half_life * 4) if cutoff else lifetime

def _default_treatment(html):
    """The default treatment for any url that doesn't match any key in custom_treatments.
    Returns a tuple made of a timeout provided by random_radioactive_lifetime and the html
    without any scripts nor styles, to save space"""

    return (_time.time() + _random_radioactive_lifetime(), _utils.trim_html(html))

def _cache_entry_is_valid(key):
    """Returns True if the cache entry is valid, otherwise returns false"""
    global cache
    return cache[key][0] > _time.time()

def get_html(url, params = {}, use_cache = True):
    """More or less functionally equivalent to ``requests.get(url, params).text``, with the added
    benefit of a persistent cache with customizable html treatment and timeouts, depending on the url.
    If the result is already in cache and is valid, the function will just return the value from the
    cache instead of making a web request.
    
    Args:
        url (str): The url of the html to be fetched
        params (:obj:`dict`, optional): the query portion of the url, should you want to include a query
        use_cache(:obj:`bool`, optional): If this value is set to True, the cache will be checked
            for the url. If the url is not found in the cache keys or has timed out, the function
            will get the html from the web, remove scripts and styles from the html, store it in cache,
            and finally return the html. Otherwise, if it's set to False, the cache will not be checked

    Returns:
        A string which is the html from the requested page url

    Note:
        The curricular units' pages, along with the students' and teachers' htmls,
        are modified to reduce their memory footprint.

    Note: 
        If you know that you are going to make a crapton of requests beforehand, you probably
        should call :func:`get_html_async` first to populate the cache.
    """
    global cache

    if cache == None:
        load_cache()
    
    if params != {}:
        url = url + "?" + _urllib.parse.urlencode(params, doseq = True) # Emulating the RequestsIII library
    
    if url in cache and _cache_entry_is_valid(url) and use_cache:
        pass
    else: # We need to get the page from the web
        request = _requests.get(url) # Getting the webpage
        request.raise_for_status()  # If the request fails, I want to know about it
        html = request.text

        for match_rule, custom_treatment in _custom_treatments.items(): # Can we apply any custom treatment to the html?
            if match_rule(url):
                cache[url] = custom_treatment(html)
                break
        else:                                                         # If not, use the default treatment
            cache[url] = _default_treatment(html)
    
    return cache[url][1]

def reset():
    """Eliminates all entries from the cache"""
    global cache

    if cache == None:
        load_cache()

    cache.clear()

def remove_invalid_entries(urls = None):
    """Removes all the cache entries in urls that have timed out.
    
    Args:
        urls (:obj:`iterable(str)` or :obj:`None`, optional): The urls to be checked. If this argument
            is left untouched, all urls in the cache will be checked
    
    """
    global cache

    if cache == None:
        load_cache()

    if urls == None:
        urls = cache.keys()
    
    hit_list = set() # Each element must be unique, otherwise it will be popped twice

    for url in urls:
        if url in cache and not _cache_entry_is_valid(url):
            hit_list.add(url)
    
    for url in hit_list:
        cache.pop(url)

def get_html_async(urls, n_workers = 10, use_cache = True):
    """:func:`get_html`, but async, give or take.

    Takes a list (or any iterable) of urls and returns a corresponding generator of htmls.
    The htmls have their scripts and styles removed and are stored in cache.

    Args:
        urls (iterable(str)): The urls to be accessed
        n_workers (:obj:`int`, optional): The number of workers.
        use_cache (:obj:`bool`, optional): Attempts to use the cache if True, otherwise it will fetch from sigarra
        
    Returns:
      An str generator
    """
    global cache

    if cache == None:
        load_cache()

    urls = tuple(urls)

    remove_invalid_entries(urls)

    work_queue = [url for url in urls if url not in cache or not use_cache]

    if len(work_queue) > 0:
        with _sessions.FuturesSession(max_workers = n_workers) as session:
            
            futures = [session.get(url) for url in work_queue]

            for future in futures:
                request = future.result()

                if request.status_code != 200:
                    continue # GTFO
                
                url  = request.url
                html = request.text

                for match_rule, custom_treatment in _custom_treatments.items(): # Can we apply any custom treatment to the html?
                    if match_rule(url):
                        cache[url] = custom_treatment(html)
                        break
                else:                                                         # If not, use the default treatment
                    cache[url] = _default_treatment(html)
    
    return (get_html(url) for url in urls)



# https://packaging.python.org/tutorials/packaging-projects/
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setuptools.setup(
    name="feupy",
    version="0.4.4",
    author="Daniel Monteiro",
    author_email="up201806185@fe.up.pt",
    description="The sigarra scraping library no one asked for",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/up201806185/feupy",

    python_requires='>=3.6',
    install_requires=required,
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
)